from datetime import date

import numpy as np
import pandas as pd
from wbcore.contrib.geography.models import Geography

REPORT_TYPE_MAP = {1: "Q", 2: "6M", 3: "4M", 4: "Y"}


def get_exchange(code):
    return {"mic_code": code}


def get_country(code):
    country = Geography.countries.filter(code_2=code).first()
    return country.id if country else None


def _clean_and_return_dict(df, extra_normalization_map=None):
    # df = df.groupby(df.columns, axis=1).sum()
    if extra_normalization_map:
        for col, den in extra_normalization_map.items():
            if col in df.columns:
                df[col] *= den
    if percent_fields_columns := list(filter(lambda x: "margin" in x, df.columns)):
        df[percent_fields_columns] = (
            df[percent_fields_columns] / 100
        )  # margin/percent type of data are given in base 100 by refinitiv
    df["instrument"] = df["instrument"].str.replace("^<|>", "", regex=True)
    df = df.rename(columns={"instrument": "instrument__provider_id"})
    # df["instrument__refinitiv_identifier_code"] = df["instrument__provider_id"] # We add this for backward compatibility reason
    df = df.replace([np.inf, -np.inf, np.nan], None)
    return df.to_dict("records")


def parse_daily_fundamental_data(import_source, default_mapping, extra_normalization_map=None):
    df = pd.read_json(import_source.file, orient="records")
    df = df.replace([np.inf, -np.inf, np.nan], None)
    mapping = {**default_mapping, "Instrument": "instrument", "Dates": "date"}
    df = df.rename(columns=mapping).dropna(how="all", axis=1)
    df = df.drop(columns=df.columns.difference(mapping.values()))
    data = list()
    if not df.empty:
        df["date"] = pd.to_datetime(df["date"], utc=True, unit="ms")
        df = df.sort_values(by="date", ascending=True)
        df.date = df.date.dt.strftime("%Y-%m-%d")
        data = _clean_and_return_dict(df, extra_normalization_map=extra_normalization_map)
    return {"data": data}


def parse_periodic_fundamental_data(import_source, default_mapping, extra_normalization_map=None):
    def _get_fiscal_period(_date, frequency):
        fiscal_period = 0
        for index, p in enumerate(
            pd.period_range(start=date(_date.year, 1, 1), end=date(_date.year, 12, 31), freq=frequency)
        ):
            if _date >= p.start_time.date() and _date <= p.end_time.date():
                fiscal_period = index + 1
        return fiscal_period

    df = pd.read_json(import_source.file, orient="records")
    df = df.replace([np.inf, -np.inf, np.nan], None)
    mapping = {**default_mapping, "Instrument": "instrument", "Dates": "date"}
    df = df.rename(columns=mapping).dropna(how="all", axis=1)

    df = df.drop(columns=df.columns.difference([*mapping.values(), "period__period_interim"]))
    df = df.dropna(
        how="all",
        subset=df.columns.difference(
            ["instrument", "date", "period__period_interim", "period__period_end_date", "period__period_type"]
        ),
    )
    df = df.dropna(
        how="any",
        subset=df.columns.intersection(["instrument", "date", "period__period_end_date", "period__period_type"]),
    )
    data = list()
    if not df.empty:
        df["date"] = pd.to_datetime(df["date"], utc=True, unit="ms")
        df["period__period_end_date"] = pd.to_datetime(df["period__period_end_date"], utc=True)
        df["period__period_year"] = df.date.dt.year
        df["period__period_type"] = df["period__period_type"].map(REPORT_TYPE_MAP)
        df["period__period_type"] = df["period__period_type"].fillna("Q")  # Assuming empty period type is quaterly
        df["period__period_index"] = df[["date", "period__period_type"]].apply(
            lambda x: _get_fiscal_period(x["date"].date(), x["period__period_type"]), axis=1
        )
        del df["date"]

        df["period__period_end_date"] = df["period__period_end_date"] + pd.offsets.MonthEnd(0)
        df["period__period_end_date"] = df["period__period_end_date"].dt.strftime("%Y-%m-%d")
        df = df.sort_values(by="period__period_end_date")

        if "employee_count" in df.columns:
            df["employee_count"] = df["employee_count"].bfill()
        data = _clean_and_return_dict(df, extra_normalization_map=extra_normalization_map)
    return {"data": data}



from contextlib import suppress
from datetime import date
from typing import Any

from wbfdm.contrib.metric.signals import instrument_metric_updated


def compute_metrics(val_date: date, key: str | None = None, basket: Any | None = None, **kwargs):
    """
    Compute and process metrics for a given date using the MetricOrchestrator.

    Args:
        val_date (date): The validation date for the metrics computation.
        key (Optional[str]): The optional metric backend key to narrow down the set of backends to use. Defaults to None.
        basket (Optional[Any]): An optional basket to narrow down the backend queryset. Defaults to None.
        **kwargs: Additional keyword arguments to pass to the MetricOrchestrator.

    Returns:
        None
    """
    from wbfdm.contrib.metric.orchestrators import MetricOrchestrator

    with suppress(KeyError):
        orchestrator = MetricOrchestrator(val_date, key=key, basket=basket, **kwargs)
        orchestrator.process()
        instrument_metric_updated.send(sender=basket.__class__, basket=basket, key=key, val_date=val_date)



import wbcore.serializers as wb_serializer
from wbcore.serializers.fields.number import DisplayMode


def get_default_attrs(cls, field_name):
    field = cls._meta.get_field(field_name)
    return {"label": field.verbose_name, "help_text": field.help_text}


class FundamentalSerializerFieldMixin:
    @classmethod
    def get_number_serializer_fields(cls):
        return {
            "revenue": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "revenue"),
            ),
            "revenue_growth": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "revenue_growth")
            ),
            "cost_of_good_sold_without_depreciation": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "cost_of_good_sold_without_depreciation"),
            ),
            "cost_of_good_sold": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "cost_of_good_sold"),
            ),
            "gross_profit": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "gross_profit"),
            ),
            "gross_profit_margin": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "gross_profit_margin")
            ),
            "eps": wb_serializer.FloatField(
                precision=2,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "eps"),
            ),
            "eps_growth": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "eps_growth")
            ),
            "diluted_eps": wb_serializer.FloatField(
                precision=2,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "diluted_eps"),
            ),
            "ebitda": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "ebitda"),
            ),
            "ebitda_margin": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "ebitda_margin")
            ),
            "ebit": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "ebit"),
            ),
            "ebit_margin": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "ebit_margin")
            ),
            "net_profit": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "net_profit"),
            ),
            "net_profit_margin": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "net_profit_margin")
            ),
            "company_tax_rate": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "company_tax_rate")
            ),
            "cost_research_development": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "cost_research_development"),
            ),
            "interest_expense": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "interest_expense"),
            ),
            "interest_expense_to_sales_ratio": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "interest_expense_to_sales_ratio")
            ),
            "sga": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "sga"),
            ),
            "sga_to_sales_ratio": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "sga_to_sales_ratio")
            ),
            "deprecation_and_amortization": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "deprecation_and_amortization"),
            ),
            "deprecation_and_amortization_to_sales_ratio": wb_serializer.FloatField(
                percent=True,
                precision=3,
                read_only=True,
                **get_default_attrs(cls, "deprecation_and_amortization_to_sales_ratio"),
            ),
            "free_cash": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "free_cash"),
            ),
            "free_cash_flow": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "free_cash_flow"),
            ),
            "free_cash_flow_growth": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "free_cash_flow_growth")
            ),
            "free_cash_flow_to_sales_ratio": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "free_cash_flow_to_sales_ratio")
            ),
            "cash_from_operation": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "cash_from_operation"),
            ),
            "working_capital": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "working_capital"),
            ),
            "capital_expenditures": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "capital_expenditures"),
            ),
            "investment_cash": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "investment_cash"),
            ),
            "financing_cash": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "financing_cash"),
            ),
            "shareholder_equity": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "shareholder_equity"),
            ),
            "total_assets": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "total_assets"),
            ),
            "current_liabilities": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "current_liabilities"),
            ),
            "total_liabilities": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "total_liabilities"),
            ),
            "total_debt": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "total_debt"),
            ),
            "cash_and_cash_equivalents": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "cash_and_cash_equivalents"),
            ),
            "net_debt": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "net_debt"),
            ),
            "cash_and_short_term_investments": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "cash_and_short_term_investments"),
            ),
            "net_change_in_cash": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "net_change_in_cash"),
            ),
            "receivables": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "receivables"),
            ),
            "inventories": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "inventories"),
            ),
            "payables": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "payables"),
            ),
            "book_value_per_share": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "book_value_per_share"),
            ),
            "current_assets": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "current_assets"),
            ),
            "employee_count": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                read_only=True,
                **get_default_attrs(cls, "employee_count"),
            ),
            "employee_count_growth": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "employee_count_growth")
            ),
            "entreprise_value": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "entreprise_value"),
            ),
            "net_debt_to_ebitda_ratio": wb_serializer.FloatField(
                precision=2, read_only=True, **get_default_attrs(cls, "net_debt_to_ebitda_ratio")
            ),
            "burn_rate": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "burn_rate"),
            ),
            "operating_burn_rate": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "operating_burn_rate"),
            ),
            "free_cash_burn_rate": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "free_cash_burn_rate"),
            ),
            "cash_reserve_to_burn_rate_ratio": wb_serializer.FloatField(
                precision=1, read_only=True, **get_default_attrs(cls, "cash_reserve_to_burn_rate_ratio")
            ),
            "cash_reserve_to_operating_burn_rate_ratio": wb_serializer.FloatField(
                precision=1, read_only=True, **get_default_attrs(cls, "cash_reserve_to_operating_burn_rate_ratio")
            ),
            "cash_reserve_to_free_cash_burn_rate_ratio": wb_serializer.FloatField(
                precision=1, read_only=True, **get_default_attrs(cls, "cash_reserve_to_free_cash_burn_rate_ratio")
            ),
            "working_capital_to_burn_rate_ratio": wb_serializer.FloatField(
                precision=1, read_only=True, **get_default_attrs(cls, "working_capital_to_burn_rate_ratio")
            ),
            "working_capital_to_operating_burn_rate_ratio": wb_serializer.FloatField(
                precision=1, read_only=True, **get_default_attrs(cls, "working_capital_to_operating_burn_rate_ratio")
            ),
            "working_capital_to_free_cash_burn_rate_ratio": wb_serializer.FloatField(
                precision=1, read_only=True, **get_default_attrs(cls, "working_capital_to_free_cash_burn_rate_ratio")
            ),
            "current_ratio": wb_serializer.FloatField(
                precision=1, read_only=True, **get_default_attrs(cls, "current_ratio")
            ),
            "cash_and_short_term_investments_to_current_assets_ratio": wb_serializer.FloatField(
                precision=1,
                read_only=True,
                **get_default_attrs(cls, "cash_and_short_term_investments_to_current_assets_ratio"),
            ),
            "rd_to_sales_ratio": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "rd_to_sales_ratio")
            ),
            "interest_coverage_ratio": wb_serializer.FloatField(
                precision=1, read_only=True, **get_default_attrs(cls, "interest_coverage_ratio")
            ),
            "return_on_equity": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "return_on_equity")
            ),
            "return_on_assets": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "return_on_assets")
            ),
            "return_on_capital_employed": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "return_on_capital_employed")
            ),
            "return_on_invested_capital": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "return_on_invested_capital")
            ),
            "revenue_to_employee_ratio": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "revenue_to_employee_ratio"),
            ),
            "capex_to_sales": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "capex_to_sales")
            ),
        }


class GeographicSegmentSerializerFieldMixin:
    @classmethod
    def get_number_serializer_fields(cls):
        return {
            "value": wb_serializer.FloatField(read_only=True, **get_default_attrs(cls, "value")),
            "value_growth": wb_serializer.FloatField(read_only=True, **get_default_attrs(cls, "value_growth")),
        }


class DailyFundamentalSerializerFieldMixin:
    @classmethod
    def get_number_serializer_fields(cls):
        return {
            "eps_ttm": wb_serializer.FloatField(
                precision=2,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "eps_ttm"),
            ),
            "eps_ftw": wb_serializer.FloatField(
                precision=2,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "eps_ftw"),
            ),
            "free_cash": wb_serializer.FloatField(
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "free_cash"),
            ),
            "free_cash_flow": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "free_cash_flow"),
            ),
            "free_cash_flow_ttm_growth": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "free_cash_flow_ttm_growth")
            ),
            "revenue_growth_3y_cagr": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "revenue_growth_3y_cagr")
            ),
            "eps_3y_cagr": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "eps_3y_cagr")
            ),
            "free_cash_flow_3y_cagr": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "free_cash_flow_3y_cagr")
            ),
        }


class ForecastSerializerFieldMixin:
    @classmethod
    def get_number_serializer_fields(cls):
        return {
            "revenue_y1": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "revenue_y1"),
            ),
            "revenue_y2": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "revenue_y2"),
            ),
            "revenue_y3": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "revenue_y3"),
            ),
            "revenue_y4": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "revenue_y4"),
            ),
            "revenue_y5": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "revenue_y5"),
            ),
            "revenue_growth_y1": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "revenue_growth_y1")
            ),
            "revenue_growth_y2": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "revenue_growth_y2")
            ),
            "revenue_growth_y3": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "revenue_growth_y3")
            ),
            "revenue_growth_y4": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "revenue_growth_y4")
            ),
            "revenue_growth_y5": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "revenue_growth_y5")
            ),
            "deprecation_and_amortization_y1": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "deprecation_and_amortization_y1"),
            ),
            "deprecation_and_amortization_y2": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "deprecation_and_amortization_y2"),
            ),
            "deprecation_and_amortization_y3": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "deprecation_and_amortization_y3"),
            ),
            "deprecation_and_amortization_y4": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "deprecation_and_amortization_y4"),
            ),
            "deprecation_and_amortization_y5": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "deprecation_and_amortization_y5"),
            ),
            "gross_profit_margin_without_depreciation_y1": wb_serializer.FloatField(
                percent=True,
                precision=3,
                read_only=True,
                **get_default_attrs(cls, "gross_profit_margin_without_depreciation_y1"),
            ),
            "gross_profit_margin_without_depreciation_y2": wb_serializer.FloatField(
                percent=True,
                precision=3,
                read_only=True,
                **get_default_attrs(cls, "gross_profit_margin_without_depreciation_y2"),
            ),
            "gross_profit_margin_without_depreciation_y3": wb_serializer.FloatField(
                percent=True,
                precision=3,
                read_only=True,
                **get_default_attrs(cls, "gross_profit_margin_without_depreciation_y3"),
            ),
            "gross_profit_margin_without_depreciation_y4": wb_serializer.FloatField(
                percent=True,
                precision=3,
                read_only=True,
                **get_default_attrs(cls, "gross_profit_margin_without_depreciation_y4"),
            ),
            "gross_profit_margin_without_depreciation_y5": wb_serializer.FloatField(
                percent=True,
                precision=3,
                read_only=True,
                **get_default_attrs(cls, "gross_profit_margin_without_depreciation_y5"),
            ),
            "gross_profit_margin_y1": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "gross_profit_margin_y1")
            ),
            "gross_profit_margin_y2": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "gross_profit_margin_y2")
            ),
            "gross_profit_margin_y3": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "gross_profit_margin_y3")
            ),
            "gross_profit_margin_y4": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "gross_profit_margin_y4")
            ),
            "gross_profit_margin_y5": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "gross_profit_margin_y5")
            ),
            "gross_profit_y1": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "gross_profit_y1"),
            ),
            "gross_profit_y2": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "gross_profit_y2"),
            ),
            "gross_profit_y3": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "gross_profit_y3"),
            ),
            "gross_profit_y4": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "gross_profit_y4"),
            ),
            "gross_profit_y5": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "gross_profit_y5"),
            ),
            "reported_net_profit_y1": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "reported_net_profit_y1"),
            ),
            "reported_net_profit_y2": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "reported_net_profit_y2"),
            ),
            "reported_net_profit_y3": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "reported_net_profit_y3"),
            ),
            "reported_net_profit_y4": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "reported_net_profit_y4"),
            ),
            "reported_net_profit_y5": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "reported_net_profit_y5"),
            ),
            "adjusted_net_profit_y1": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "adjusted_net_profit_y1"),
            ),
            "adjusted_net_profit_y2": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "adjusted_net_profit_y2"),
            ),
            "adjusted_net_profit_y3": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "adjusted_net_profit_y3"),
            ),
            "adjusted_net_profit_y4": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "adjusted_net_profit_y4"),
            ),
            "adjusted_net_profit_y5": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "adjusted_net_profit_y5"),
            ),
            "net_profit_margin_y1": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "net_profit_margin_y1")
            ),
            "net_profit_margin_y2": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "net_profit_margin_y2")
            ),
            "net_profit_margin_y3": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "net_profit_margin_y3")
            ),
            "net_profit_margin_y4": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "net_profit_margin_y4")
            ),
            "net_profit_margin_y5": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "net_profit_margin_y5")
            ),
            "ebitda_y1": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "ebitda_y1"),
            ),
            "ebitda_y2": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "ebitda_y2"),
            ),
            "ebitda_y3": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "ebitda_y3"),
            ),
            "ebitda_y4": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "ebitda_y4"),
            ),
            "ebitda_y5": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "ebitda_y5"),
            ),
            "ebitda_margin_y1": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "ebitda_margin_y1")
            ),
            "ebitda_margin_y2": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "ebitda_margin_y2")
            ),
            "ebitda_margin_y3": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "ebitda_margin_y3")
            ),
            "ebitda_margin_y4": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "ebitda_margin_y4")
            ),
            "ebitda_margin_y5": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "ebitda_margin_y5")
            ),
            "ebit_y1": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "ebit_y1"),
            ),
            "ebit_y2": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "ebit_y2"),
            ),
            "ebit_y3": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "ebit_y3"),
            ),
            "ebit_y4": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "ebit_y4"),
            ),
            "ebit_y5": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "ebit_y5"),
            ),
            "ebit_margin_y1": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "ebit_margin_y1")
            ),
            "ebit_margin_y2": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "ebit_margin_y2")
            ),
            "ebit_margin_y3": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "ebit_margin_y3")
            ),
            "ebit_margin_y4": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "ebit_margin_y4")
            ),
            "ebit_margin_y5": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "ebit_margin_y5")
            ),
            "eps_y1": wb_serializer.FloatField(
                precision=2,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "eps_y1"),
            ),
            "eps_y2": wb_serializer.FloatField(
                precision=2,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "eps_y2"),
            ),
            "eps_y3": wb_serializer.FloatField(
                precision=2,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "eps_y3"),
            ),
            "eps_y4": wb_serializer.FloatField(
                precision=2,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "eps_y4"),
            ),
            "eps_y5": wb_serializer.FloatField(
                precision=2,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "eps_y5"),
            ),
            "eps_growth_y1": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "eps_growth_y1")
            ),
            "eps_growth_y2": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "eps_growth_y2")
            ),
            "eps_growth_y3": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "eps_growth_y3")
            ),
            "eps_growth_y4": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "eps_growth_y4")
            ),
            "eps_growth_y5": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "eps_growth_y5"),
            ),
            "net_debt_y1": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "net_debt_y1"),
            ),
            "net_debt_y2": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "net_debt_y2"),
            ),
            "net_debt_y3": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "net_debt_y3"),
            ),
            "net_debt_y4": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "net_debt_y4"),
            ),
            "net_debt_y5": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "net_debt_y5"),
            ),
            "net_debt_to_ebitda_ratio_y1": wb_serializer.FloatField(
                precision=1, read_only=True, **get_default_attrs(cls, "net_debt_to_ebitda_ratio_y1")
            ),
            "net_debt_to_ebitda_ratio_y2": wb_serializer.FloatField(
                precision=1, read_only=True, **get_default_attrs(cls, "net_debt_to_ebitda_ratio_y2")
            ),
            "net_debt_to_ebitda_ratio_y3": wb_serializer.FloatField(
                precision=1, read_only=True, **get_default_attrs(cls, "net_debt_to_ebitda_ratio_y3")
            ),
            "net_debt_to_ebitda_ratio_y4": wb_serializer.FloatField(
                precision=1, read_only=True, **get_default_attrs(cls, "net_debt_to_ebitda_ratio_y4")
            ),
            "net_debt_to_ebitda_ratio_y5": wb_serializer.FloatField(
                precision=1, read_only=True, **get_default_attrs(cls, "net_debt_to_ebitda_ratio_y5")
            ),
            "entreprise_value_y1": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "entreprise_value_y1"),
            ),
            "entreprise_value_y2": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "entreprise_value_y2"),
            ),
            "entreprise_value_y3": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "entreprise_value_y3"),
            ),
            "entreprise_value_y4": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "entreprise_value_y4"),
            ),
            "entreprise_value_y5": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "entreprise_value_y5"),
            ),
            "free_cash_flow_y1": wb_serializer.FloatField(
                precision=2,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "free_cash_flow_y1"),
            ),
            "free_cash_flow_y2": wb_serializer.FloatField(
                precision=2,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "free_cash_flow_y2"),
            ),
            "free_cash_flow_y3": wb_serializer.FloatField(
                precision=2,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "free_cash_flow_y3"),
            ),
            "free_cash_flow_y4": wb_serializer.FloatField(
                precision=2,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "free_cash_flow_y4"),
            ),
            "free_cash_flow_y5": wb_serializer.FloatField(
                precision=2,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "free_cash_flow_y5"),
            ),
            "free_cash_flow_growth_y1": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "free_cash_flow_growth_y1")
            ),
            "free_cash_flow_growth_y2": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "free_cash_flow_growth_y2")
            ),
            "free_cash_flow_growth_y3": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "free_cash_flow_growth_y3")
            ),
            "free_cash_flow_growth_y4": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "free_cash_flow_growth_y4")
            ),
            "free_cash_flow_growth_y5": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "free_cash_flow_growth_y5")
            ),
            "free_cash_flow_to_sales_ratio_y1": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "free_cash_flow_to_sales_ratio_y1")
            ),
            "free_cash_flow_to_sales_ratio_y2": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "free_cash_flow_to_sales_ratio_y2")
            ),
            "free_cash_flow_to_sales_ratio_y3": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "free_cash_flow_to_sales_ratio_y3")
            ),
            "free_cash_flow_to_sales_ratio_y4": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "free_cash_flow_to_sales_ratio_y4")
            ),
            "free_cash_flow_to_sales_ratio_y5": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "free_cash_flow_to_sales_ratio_y5")
            ),
            "expected_book_value_per_share_y1": wb_serializer.FloatField(
                precision=2,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "expected_book_value_per_share_y1"),
            ),
            "expected_book_value_per_share_y2": wb_serializer.FloatField(
                precision=2,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "expected_book_value_per_share_y2"),
            ),
            "expected_book_value_per_share_y3": wb_serializer.FloatField(
                precision=2,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "expected_book_value_per_share_y3"),
            ),
            "expected_book_value_per_share_y4": wb_serializer.FloatField(
                precision=2,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "expected_book_value_per_share_y4"),
            ),
            "expected_book_value_per_share_y5": wb_serializer.FloatField(
                precision=2,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "expected_book_value_per_share_y5"),
            ),
            "capital_expenditures_y1": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "capital_expenditures_y1"),
            ),
            "capital_expenditures_y2": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "capital_expenditures_y2"),
            ),
            "capital_expenditures_y3": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "capital_expenditures_y3"),
            ),
            "capital_expenditures_y4": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "capital_expenditures_y4"),
            ),
            "capital_expenditures_y5": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=1,
                decorators=[
                    wb_serializer.decorator(decorator_type="text", position="left", value="{{currency_repr}}")
                ],
                read_only=True,
                **get_default_attrs(cls, "capital_expenditures_y5"),
            ),
            "capex_to_sales_y1": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "capex_to_sales_y1")
            ),
            "capex_to_sales_y2": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "capex_to_sales_y2")
            ),
            "capex_to_sales_y3": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "capex_to_sales_y3")
            ),
            "capex_to_sales_y4": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "capex_to_sales_y4")
            ),
            "capex_to_sales_y5": wb_serializer.FloatField(
                percent=True, precision=3, read_only=True, **get_default_attrs(cls, "capex_to_sales_y5")
            ),
            "return_on_equity_y1": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_equity_y1")
            ),
            "return_on_equity_y2": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_equity_y2")
            ),
            "return_on_equity_y3": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_equity_y3")
            ),
            "return_on_equity_y4": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_equity_y4")
            ),
            "return_on_equity_y5": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_equity_y5")
            ),
            "return_on_assets_y1": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_assets_y1")
            ),
            "return_on_assets_y2": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_assets_y2")
            ),
            "return_on_assets_y3": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_assets_y3")
            ),
            "return_on_assets_y4": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_assets_y4")
            ),
            "return_on_assets_y5": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_assets_y5")
            ),
            "return_on_capital_employed_y1": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_capital_employed_y1")
            ),
            "return_on_capital_employed_y2": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_capital_employed_y2")
            ),
            "return_on_capital_employed_y3": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_capital_employed_y3")
            ),
            "return_on_capital_employed_y4": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_capital_employed_y4")
            ),
            "return_on_capital_employed_y5": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_capital_employed_y5")
            ),
            "return_on_invested_capital_y1": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_invested_capital_y1")
            ),
            "return_on_invested_capital_y2": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_invested_capital_y2")
            ),
            "return_on_invested_capital_y3": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_invested_capital_y3")
            ),
            "return_on_invested_capital_y4": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_invested_capital_y4")
            ),
            "return_on_invested_capital_y5": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "return_on_invested_capital_y5")
            ),
            "interest_coverage_ratio_y1": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "interest_coverage_ratio_y1")
            ),
            "interest_coverage_ratio_y2": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "interest_coverage_ratio_y2")
            ),
            "interest_coverage_ratio_y3": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "interest_coverage_ratio_y3")
            ),
            "interest_coverage_ratio_y4": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "interest_coverage_ratio_y4")
            ),
            "interest_coverage_ratio_y5": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "interest_coverage_ratio_y5")
            ),
        }


class InstrumentPriceSerializerFieldMixin:
    @classmethod
    def get_number_serializer_fields(cls):
        return {
            "net_value": wb_serializer.FloatField(precision=2, read_only=True, **get_default_attrs(cls, "net_value")),
            "gross_value": wb_serializer.FloatField(
                precision=2, read_only=True, **get_default_attrs(cls, "gross_value")
            ),
            "outstanding_shares": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED,
                precision=2,
                read_only=True,
                **get_default_attrs(cls, "outstanding_shares"),
            ),
            "outstanding_shares_consolidated": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "outstanding_shares_consolidated")
            ),
            "volume": wb_serializer.FloatField(
                display_mode=DisplayMode.SHORTENED, precision=2, read_only=True, **get_default_attrs(cls, "volume")
            ),
            "volume_50d": wb_serializer.FloatField(read_only=True, **get_default_attrs(cls, "volume_50d")),
            "volume_200d": wb_serializer.FloatField(read_only=True, **get_default_attrs(cls, "volume_200d")),
            "market_capitalization": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "market_capitalization")
            ),
            "market_capitalization_consolidated": wb_serializer.FloatField(
                read_only=True, **get_default_attrs(cls, "market_capitalization_consolidated")
            ),
            "sharpe_ratio": wb_serializer.FloatField(read_only=True, **get_default_attrs(cls, "sharpe_ratio")),
            "correlation": wb_serializer.FloatField(read_only=True, **get_default_attrs(cls, "correlation")),
            "beta": wb_serializer.FloatField(read_only=True, **get_default_attrs(cls, "beta")),
            # "custom_beta_180d": wb_serializer.FloatField(read_only=True, **get_default_attrs(cls, "custom_beta_180d")),
            # "custom_beta_1y": wb_serializer.FloatField(read_only=True, **get_default_attrs(cls, "custom_beta_1y")),
            # "custom_beta_2y": wb_serializer.FloatField(read_only=True, **get_default_attrs(cls, "custom_beta_2y")),
            # "custom_beta_3y": wb_serializer.FloatField(read_only=True, **get_default_attrs(cls, "custom_beta_3y")),
            # "custom_beta_5y": wb_serializer.FloatField(read_only=True, **get_default_attrs(cls, "custom_beta_5y")),
            # "performance_1d": wb_serializer.FloatField(read_only=True, **get_default_attrs(cls, "performance_1d")),
            # "performance_7d": wb_serializer.FloatField(read_only=True, **get_default_attrs(cls, "performance_7d")),
            # "performance_30d": wb_serializer.FloatField(read_only=True, **get_default_attrs(cls, "performance_30d")),
            # "performance_90d": wb_serializer.FloatField(read_only=True, **get_default_attrs(cls, "performance_90d")),
            # "performance_365d": wb_serializer.FloatField(read_only=True, **get_default_attrs(cls, "performance_365d")),
            # "performance_ytd": wb_serializer.FloatField(read_only=True, **get_default_attrs(cls, "performance_ytd")),
            # "performance_inception": wb_serializer.FloatField(
            #     read_only=True, **get_default_attrs(cls, "performance_inception")
            # ),
        }



from django.contrib import admin

from wbfdm.models import InstrumentRequest


@admin.register(InstrumentRequest)
class InstrumentRequestModelAdmin(admin.ModelAdmin):
    search_fields = (
        "created_instrument__computed_str",
        "requester__comptuted_str",
    )
    raw_id_fields = ["created_instrument", "requester"]
    list_display = (
        "status",
        "requester",
        "handler",
        "notes",
        "created",
        "created_instrument",
    )
    readonly_fields = ["created"]
    fieldsets = (
        (
            "Main Data",
            {
                "fields": (
                    ("status", "created", "created_instrument"),
                    ("requester", "handler"),
                    ("notes",),
                    ("instrument_data",),
                )
            },
        ),
    )



from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

INSTRUMENT_LIST_MENUITEM = MenuItem(
    label="Instrument Lists",
    endpoint="wbfdm:instrumentlist-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["wbfdm.view_instrumentlist"]
    ),
)



from typing import Any, Self

from django.db import models

from wbfdm.enums import (
    ESGControveryFlag,
    ESGControverySeverity,
    ESGControveryStatus,
    ESGControveryType,
)


class Controversy(models.Model):
    external_id = models.CharField(max_length=512, unique=True)
    instrument = models.ForeignKey("wbfdm.Instrument", on_delete=models.CASCADE, limit_choices_to=models.Q(level=0))
    headline = models.TextField(verbose_name="Headline")
    description = models.TextField(verbose_name="Description")
    source = models.TextField(
        verbose_name="Source", null=True, blank=True
    )  # Source is usually a sentence but can be a whole text
    status = models.CharField(
        max_length=64,
        choices=ESGControveryStatus.choices,
        verbose_name="Status",
        default=ESGControveryStatus.ONGOING.value,
        null=True,
        blank=True,
    )
    type = models.CharField(
        max_length=64,
        choices=ESGControveryType.choices,
        verbose_name="Type",
        default=ESGControveryType.STRUCTURAL.value,
        null=True,
        blank=True,
    )
    severity = models.CharField(
        max_length=64,
        choices=ESGControverySeverity.choices,
        verbose_name="Severity",
        default=ESGControverySeverity.MINOR.value,
        null=True,
        blank=True,
    )
    flag = models.CharField(
        max_length=64,
        choices=ESGControveryFlag.choices,
        verbose_name="Flag",
        default=ESGControveryFlag.GREEN.value,
        null=True,
        blank=True,
    )
    direct_involvement = models.BooleanField(default=True, verbose_name="Direct Involvement")
    company_response = models.CharField(max_length=512, null=True, blank=True)

    review = models.DateField(verbose_name="Reviewed", null=True, blank=True)
    initiated = models.DateField(verbose_name="initiated", null=True, blank=True)

    def __str__(self) -> str:
        return f"{self.headline} ({self.instrument})"

    @classmethod
    def dict_to_model(cls, controversy: dict[str, Any], instrument) -> Self:
        return Controversy(
            external_id=controversy["id"],
            instrument=instrument,
            headline=controversy["headline"],
            description=controversy["narrative"],
            source=controversy["source"],
            direct_involvement=controversy.get("direct_involvement", True),
            company_response=controversy["response"],
            review=controversy["review"],
            initiated=controversy["initiated"],
            flag=controversy["flag"],
            status=controversy["status"],
            type=controversy["type"],
            severity=controversy["assessment"],
        )



import logging
import re
from contextlib import suppress
from datetime import date, timedelta
from typing import Any, Generator, Iterator, Self, TypeVar

import pandas as pd
from celery import shared_task
from colorfield.fields import ColorField
from django.contrib import admin
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields import ArrayField
from django.contrib.postgres.indexes import GinIndex
from django.contrib.postgres.search import SearchVector, SearchVectorField
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.db import models, transaction
from django.db.models import Q, Value
from django.db.models.signals import post_delete, post_save, pre_delete, pre_save
from django.dispatch import receiver
from dynamic_preferences.registries import global_preferences_registry
from mptt.models import MPTTModel, TreeForeignKey, TreeManager
from pandas.tseries.offsets import BDay
from rest_framework.reverse import reverse
from slugify import slugify
from tqdm import tqdm
from wbcore.content_type.utils import get_ancestors_content_type
from wbcore.contrib.dataloader.models import Entity
from wbcore.contrib.io.mixins import ImportMixin
from wbcore.contrib.io.models import ImportedObjectProviderRelationship
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.contrib.tags.models import TagModelMixin
from wbcore.models import WBModel
from wbcore.signals import pre_merge
from wbcore.utils.models import ComplexToStringMixin
from wbnews.models import News
from wbnews.signals import create_news_relationships

from wbfdm.analysis import TechnicalAnalysis
from wbfdm.contrib.internal.dataloaders.market_data import MarketDataDataloader
from wbfdm.contrib.metric.dispatch import compute_metrics
from wbfdm.import_export.handlers.instrument import InstrumentImportHandler
from wbfdm.models.instruments.llm.create_instrument_news_relationships import (
    run_company_extraction_llm,
)
from wbfdm.preferences import get_default_classification_group
from wbfdm.signals import (
    add_instrument_to_investable_universe,
    instrument_price_imported,
    investable_universe_updated,
)

from ...analysis.financial_analysis.change_point_detection import outlier_detection, statistical_change_point_detection
from ...dataloaders.proxies import InstrumentDataloaderProxy
from .instrument_relationships import RelatedInstrumentThroughModel
from .mixin.instruments import InstrumentPMSMixin
from .querysets import InstrumentQuerySet
from .utils import re_bloomberg, re_isin, re_mnemonic, re_ric

logger = logging.getLogger("pms")


class InstrumentManager(TreeManager):
    def __init__(self, with_annotation: bool = False, *args, **kwargs):
        self.with_annotation = with_annotation
        super().__init__(*args, **kwargs)

    def _custom_rebuild_helper(self, node, left, tree_id, nodes_to_update, level):
        right = left + 1

        for child in node.children.all():
            right = self._custom_rebuild_helper(
                node=child,
                left=right,
                tree_id=tree_id,
                nodes_to_update=nodes_to_update,
                level=level + 1,
            )

        setattr(node, self._rebuild_fields["left"], left)
        setattr(node, self._rebuild_fields["right"], right)
        setattr(node, self._rebuild_fields["level"], level)
        setattr(node, self._rebuild_fields["tree_id"], tree_id)
        nodes_to_update.append(node)

        return right + 1

    def rebuild(self, batch_size=1000, debug: bool = False, **filters) -> None:
        """
        We supercharge MPTT rebuild manager method to avoid loading all instrument into memory
        """
        self._find_out_rebuild_fields()

        parents = self._get_parents(**filters)
        tree_id = filters.get("tree_id", 1)
        nodes_to_update = []
        if debug:
            gen = tqdm(enumerate(parents), total=len(parents))
        else:
            gen = enumerate(parents)
        for index, parent in gen:
            self._custom_rebuild_helper(
                node=parent,
                left=1,
                tree_id=tree_id + index,
                nodes_to_update=nodes_to_update,
                level=0,
            )
            if len(nodes_to_update) >= batch_size:
                self.bulk_update(
                    nodes_to_update,
                    self._rebuild_fields.values(),
                )
                nodes_to_update = []

        self.bulk_update(
            nodes_to_update,
            self._rebuild_fields.values(),
        )

    def get_queryset(self) -> InstrumentQuerySet:
        qs = InstrumentQuerySet(self.model, using=self._db)
        if self.with_annotation:
            qs = qs.annotate_all()
        return qs

    def annotate_classification_for_group(
        self, classification_group, classification_height: int = 0, **kwargs
    ) -> models.QuerySet:
        return self.get_queryset().annotate_classification_for_group(
            classification_group, classification_height=classification_height, **kwargs
        )

    def annotate_base_data(self):
        return self.get_queryset().annotate_base_data()

    def annotate_all(self):
        return self.get_queryset().annotate_all()

    def filter_active_at_date(self, val_date: date):
        return self.get_queryset().filter_active_at_date(val_date)

    def get_instrument_prices_from_market_data(self, **kwargs):
        return self.get_queryset().get_instrument_prices_from_market_data(**kwargs)

    def get_returns_df(self, **kwargs) -> tuple[dict[date, dict[int, float]], pd.DataFrame]:
        return self.get_queryset().get_returns_df(**kwargs)


class SecurityInstrumentManager(InstrumentManager):
    def get_queryset(self) -> InstrumentQuerySet:
        return super().get_queryset().filter(is_security=True)


class ClassifiableInstrumentManager(InstrumentManager):
    def get_queryset(self) -> InstrumentQuerySet:
        return super().get_queryset().filter(instrument_type__is_classifiable=True, level=0)


class ActiveInstrumentManager(InstrumentManager):
    def get_queryset(self):
        return super().get_queryset().filter_active_at_date(date.today())


class InvestableUniverseManager(InstrumentManager):
    def get_queryset(self) -> InstrumentQuerySet:
        instrument_ids = set()
        for _, ids in add_instrument_to_investable_universe.send(sender=Instrument):
            instrument_ids.update(ids)
        return (
            super()
            .get_queryset()
            .annotate_base_data()
            .filter(is_investable=True)
            .filter(
                Q(is_investable_universe=True)
                | Q(
                    dependent_instruments_through__isnull=False
                )  # we consider instrument that are "related" to other instrument as within the investable universe by default
                | Q(id__in=instrument_ids)
                | Q(is_managed=True)
                | Q(
                    dl_parameters__market_data__path="wbfdm.contrib.internal.dataloaders.market_data.MarketDataDataloader"
                )
            )
        )


class InvestableInstrumentManager(InstrumentManager):
    def get_queryset(self) -> InstrumentQuerySet:
        return super().get_queryset().filter(children__isnull=True)


SelfInstrument = TypeVar("SelfInstrument", bound="Instrument")


class InstrumentType(models.Model):
    name = models.CharField(max_length=128, verbose_name="Name")
    short_name = models.CharField(max_length=128, verbose_name="Short Name")
    name_repr = models.CharField(max_length=128, verbose_name="Name (Representation)")
    key = models.CharField(max_length=32, verbose_name="Key", unique=True)
    description = models.TextField(verbose_name="Description", blank=True)

    is_classifiable = models.BooleanField(default=True, verbose_name="Classifiable")
    is_security = models.BooleanField(default=True, verbose_name="Security")

    @classmethod
    @property
    def PRODUCT(cls):  # noqa
        return InstrumentType.objects.get_or_create(
            key="product", defaults={"name": "Product", "short_name": "Product"}
        )[0]

    @classmethod
    @property
    def EQUITY(cls):  # noqa
        return InstrumentType.objects.get_or_create(key="equity", defaults={"name": "equity", "short_name": "equity"})[
            0
        ]

    @classmethod
    @property
    def INDEX(cls):  # noqa
        return InstrumentType.objects.get_or_create(key="index", defaults={"name": "Index", "short_name": "Index"})[0]

    @classmethod
    @property
    def CASH(cls):  # noqa
        return InstrumentType.objects.get_or_create(key="cash", defaults={"name": "Cash", "short_name": "Cash"})[0]

    @classmethod
    @property
    def CASHEQUIVALENT(cls):  # noqa
        return InstrumentType.objects.get_or_create(
            key="cash_equivalent", defaults={"name": "Cash Equivalents", "short_name": "Cash Equivalents"}
        )[0]

    @classmethod
    @property
    def PRODUCT_GROUP(cls):  # noqa
        return InstrumentType.objects.get_or_create(
            key="product_group", defaults={"name": "Product Group", "short_name": "Product Group"}
        )[0]

    def save(self, *args, **kwargs):
        if not self.short_name:
            self.short_name = self.name
        if not self.key:
            self.key = slugify(self.name, separator="_")
        if not self.name_repr:
            self.name_repr = self.name
        super().save(*args, **kwargs)

    def __str__(self) -> str:
        return f"{self.name}"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbfdm:instrumenttyperepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"


class Instrument(ComplexToStringMixin, TagModelMixin, ImportMixin, InstrumentPMSMixin, WBModel, Entity, MPTTModel):
    COMPUTED_STR_RECOMPUTE_PERIODICALLY: bool = False
    # COMPUTED_STR_RECOMPUTE_ON_SAVE: bool = False # I am commenting this out because we need computed str to be recomputed on save but do not know why this would be an issue

    import_export_handler_class = InstrumentImportHandler
    dl_proxy = InstrumentDataloaderProxy

    parent = TreeForeignKey(
        "self",
        related_name="children",
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        verbose_name="Parent Instrument",
    )
    name = models.CharField(max_length=512)
    description = models.TextField(default="", blank=True, null=True)
    instrument_type = models.ForeignKey(
        "wbfdm.InstrumentType", related_name="instruments", null=True, blank=True, on_delete=models.PROTECT
    )

    inception_date = models.DateField(null=True, blank=True)
    delisted_date = models.DateField(null=True, blank=True)
    last_valuation_date = models.DateField(
        null=True, blank=True, verbose_name="Last Valuation Date", help_text="Last Valuation Date"
    )
    last_price_date = models.DateField(
        null=True, blank=True, verbose_name="Last Price Date", help_text="Last Price Date"
    )
    # The report date fields store the actual dates when a report happens, not the end of a period.
    last_annual_report = models.DateField(null=True, blank=True)
    last_interim_report = models.DateField(null=True, blank=True)
    next_annual_report = models.DateField(null=True, blank=True)
    next_interim_report = models.DateField(null=True, blank=True)

    country = models.ForeignKey(
        to="geography.Geography",
        null=True,
        blank=True,
        limit_choices_to={"level": 1},
        on_delete=models.SET_NULL,
    )

    currency = models.ForeignKey(
        to="currency.Currency",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    exchange = models.ForeignKey(
        to="wbfdm.Exchange", null=True, blank=True, on_delete=models.PROTECT, related_name="instruments"
    )
    round_lot_size = models.IntegerField(
        default=1,
        verbose_name="Round Lot Size",
        help_text="A round lot (or board lot) is the normal unit of trading of a security.",
    )

    source_id = models.CharField(max_length=64, null=True, blank=True)
    source = models.CharField(max_length=64, null=True, blank=True)

    # Other fields from PMS
    founded_year = models.IntegerField(null=True, blank=True, verbose_name="Founded Year")
    identifier = models.CharField(
        max_length=255,
        verbose_name="Identifier",
        null=True,
        blank=True,
    )
    name_repr = models.CharField(max_length=255, null=True, blank=True, verbose_name="Name (Representation)")
    last_update = models.DateTimeField(auto_now=True, blank=True, null=True)
    alternative_names = ArrayField(models.CharField(blank=True, null=True, max_length=255), default=list, blank=True)
    isin = models.CharField(
        null=True,
        blank=True,
        max_length=12,
        verbose_name="ISIN",
        help_text="The ISIN provided by the bank.",
    )

    ticker = models.CharField(
        max_length=255,
        verbose_name="Ticker Bloomberg",
        help_text="The Bloomberg ticker without the exchange (e.g. AAPL)",
        blank=True,
        null=True,
    )

    old_isins = ArrayField(
        base_field=models.CharField(max_length=12),
        default=list,
        blank=True,
        verbose_name="Old ISINS",
        help_text="These old ISINs are stored for this instrument to retrieve it more easily later.",
    )
    sedol = models.CharField(
        max_length=255,
        verbose_name="SEDOL",
        help_text="Stock Exchange Daily Official List",
        blank=True,
        null=True,
    )

    valoren = models.CharField(
        max_length=255,
        verbose_name="Valoren Number",
        help_text="Valoren Number",
        blank=True,
        null=True,
    )

    cusip = models.CharField(
        max_length=255,
        verbose_name="CUSIP",
        help_text="CUSIP",
        blank=True,
        null=True,
    )

    refinitiv_ticker = models.CharField(
        max_length=255,
        verbose_name="Refinitiv Ticker",
        help_text="Refinitiv Refinitiv",
        blank=True,
        null=True,
    )
    refinitiv_identifier_code = models.CharField(
        max_length=255,
        verbose_name="RIC",
        help_text="Refinitiv Identifier Code",
        blank=True,
        null=True,
    )

    refinitiv_mnemonic_code = models.CharField(
        max_length=255,
        verbose_name="Refinitiv Datastream Mnemonic Code",
        help_text="Refinitiv Datastream Mnemonic Code",
        blank=True,
        null=True,
    )

    headquarter_address = models.CharField(
        max_length=512, blank=True, null=True, help_text="The company Headquarter address"
    )
    headquarter_city = models.ForeignKey(
        "geography.Geography",
        related_name="headquarters_of",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name="Headquarter City",
        help_text="The company's headquarter city",
        limit_choices_to={"level": 3},
    )
    employees = models.IntegerField(null=True, blank=True)

    primary_url = models.URLField(blank=True, null=True, help_text="The Company website url")
    additional_urls = ArrayField(models.URLField(blank=True, null=True), default=list, blank=True)

    related_instruments = models.ManyToManyField(
        "self",
        symmetrical=False,
        related_name="benchmarks_of",
        through="wbfdm.RelatedInstrumentThroughModel",
        through_fields=("instrument", "related_instrument"),
        blank=True,
        verbose_name="The Related Instruments",
    )

    classifications = models.ManyToManyField(
        "wbfdm.Classification",
        through="wbfdm.InstrumentClassificationThroughModel",
        limit_choices_to=models.Q(level=models.F("group__max_depth")),
        related_name="instruments",
        blank=True,
        verbose_name="Classifications",
    )
    is_cash = models.BooleanField(default=False)
    is_cash_equivalent = models.BooleanField(default=False)
    issue_price = models.PositiveIntegerField(
        default=100,
        verbose_name="Issue Price",
        help_text="The initial issue price that is displayed on the factsheet",
    )
    base_color = ColorField(
        blank=True, null=True, max_length=64, default="#FF0000"
    )  # we need this field for pms breakdown.

    is_security = models.BooleanField(default=False)
    is_managed = models.BooleanField(default=False)
    is_primary = models.BooleanField(null=True, blank=True)
    is_investable_universe = models.BooleanField(
        default=False,
        verbose_name="In Investable Universe",
        help_text="If True, the instrument belongs to the investable universe",
    )

    search_vector = SearchVectorField(null=True)
    trigram_search_vector = models.CharField(max_length=1024, null=True, blank=True)

    objects = InstrumentManager()
    annotated_objects = InstrumentManager(with_annotation=True)
    active_objects = ActiveInstrumentManager()
    securities = SecurityInstrumentManager()
    classifiables = ClassifiableInstrumentManager()
    investables = InvestableInstrumentManager()
    investable_universe = InvestableUniverseManager()

    class Meta:
        verbose_name = "Instrument"
        verbose_name_plural = "Instruments"
        permissions = (("administrate_instrument", "Can administrate Instrument"),)
        constraints = [
            models.UniqueConstraint(fields=["source_id", "source"], name="unique_source"),
            models.UniqueConstraint(
                fields=["refinitiv_identifier_code"],
                name="unique_ric",
                condition=models.Q(is_security=True) & models.Q(delisted_date__isnull=True),
            ),
            models.UniqueConstraint(
                fields=["refinitiv_mnemonic_code"],
                name="unique_rmc",
                condition=models.Q(is_security=True) & models.Q(delisted_date__isnull=True),
            ),
            models.UniqueConstraint(
                fields=["isin"],
                name="unique_isin",
                condition=models.Q(is_security=True) & models.Q(delisted_date__isnull=True),
            ),
            models.UniqueConstraint(
                fields=["sedol"],
                name="unique_sedol",
                condition=models.Q(is_security=True) & models.Q(delisted_date__isnull=True),
            ),
            models.UniqueConstraint(
                fields=["valoren"],
                name="unique_valoren",
                condition=models.Q(is_security=True) & models.Q(delisted_date__isnull=True),
            ),
            models.UniqueConstraint(
                fields=["cusip"],
                name="unique_cusip",
                condition=models.Q(is_security=True) & models.Q(delisted_date__isnull=True),
            ),
            models.UniqueConstraint(
                fields=["parent", "is_primary"],
                name="unique_instrument_primary",
                condition=models.Q(is_primary=True) & models.Q(is_managed=False),
            ),
        ]
        indexes = [
            models.Index(fields=["parent"], name="instrument_parent_index"),
            models.Index(fields=["parent", "exchange", "isin"], name="instrument_children_index"),
            models.Index(fields=["is_investable_universe"], name="instrument_investible_index"),
            models.Index(fields=["is_security"], name="instrument_security_index"),
            models.Index(fields=["level"], name="instrument_level_index"),
            GinIndex(fields=["search_vector"], name="instrument_sv_gin_idx"),  # type: ignore
            GinIndex(
                fields=["trigram_search_vector"], opclasses=["gin_trgm_ops"], name="instrument_trigram_sv_gin_idx"
            ),  # type: ignore
        ]
        notification_types = [
            create_notification_type(
                code="wbfdm.instrument.notify",
                title="Instrument Notification",
                help_text="Sends a notification when there is an update about an instrument",
            )
        ]

    def get_tag_detail_endpoint(self):
        return reverse("wbfdm:instrument-detail", [self.id])

    def get_tag_representation(self):
        return self.computed_str

    @property
    @admin.display(description="Is Investable")
    def _is_investable(self):
        return getattr(self, "is_investable", not self.children.exists())

    @property
    @admin.display(description="Primary Benchmark")
    def primary_benchmark(self):
        if primary_through := RelatedInstrumentThroughModel.objects.filter(
            instrument=self, is_primary=True, related_type=RelatedInstrumentThroughModel.RelatedTypeChoices.BENCHMARK
        ).first():
            return primary_through.related_instrument
        return None

    @property
    @admin.display(description="Primary Peer")
    def primary_peer(self):
        if primary_through := RelatedInstrumentThroughModel.objects.filter(
            instrument=self, is_primary=True, related_type=RelatedInstrumentThroughModel.RelatedTypeChoices.PEER
        ).first():
            return primary_through.related_instrument
        return None

    @property
    @admin.display(description="Primary Risk Instrument")
    def primary_risk_instrument(self):
        if primary_through := RelatedInstrumentThroughModel.objects.filter(
            instrument=self,
            is_primary=True,
            related_type=RelatedInstrumentThroughModel.RelatedTypeChoices.RISK_INSTRUMENT,
        ).first():
            return primary_through.related_instrument
        return None

    @property
    @admin.display(description="Primary Classification")
    def primary_classification(self):
        if primary_classification := self.classifications.filter(group__is_primary=True).first():
            return primary_classification

    @property
    @admin.display(description="Favorite Classification")
    def favorite_classification(self):
        if favorite_classification := self.classifications.filter(group=get_default_classification_group()).first():
            return favorite_classification

    def get_primary_quote(self):
        if self.children.exists():
            # we try the primary children first, then the first children within the investable universe and we fallback to the first children otherwise
            try:
                instrument = self.children.get(is_primary=True)
            except Instrument.DoesNotExist:
                instrument = self.children.filter(is_investable_universe=True).first()
                if not instrument:
                    instrument = self.children.first()
            return instrument.get_primary_quote()
        return self

    @property
    def active(self) -> bool:
        today = date.today()
        return (
            self.inception_date
            and self.inception_date <= today
            and (not self.delisted_date or self.delisted_date > today)
        )

    @property
    def identifier_repr(self) -> str:
        identifiers = []
        if self.is_security:
            identifier_labels = ["ticker", "refinitiv_identifier_code", "refinitiv_mnemonic_code"]
        else:
            identifier_labels = ["refinitiv_mnemonic_code"]
        for label in identifier_labels:
            if v := getattr(self, label, None):
                identifiers.append(v)
                break
        if self.isin and self.is_security:
            identifiers.append(self.isin)

        return " - ".join(identifiers).replace(":", "-")

    @property
    def valuations(self):
        try:
            return self.prices.filter(calculated=False)
        except (
            ValueError
        ):  # ValueError because if this property is called before the instance has a primary key, it will fail
            return Instrument.objects.none()

    @property
    def security_instrument_type(self):
        while not self.instrument_type.is_security and self.parent:
            return self.parent.security_instrument_type
        return self.instrument_type

    def update_search_vectors(self):
        names = list(map(lambda x: Value(x), filter(None, [self.name, *self.alternative_names])))
        if names:
            isins = list(map(lambda x: Value(x), filter(None, [self.isin, *self.old_isins])))
            identifiers = list(
                map(
                    lambda x: Value(re.sub(r"[^a-zA-Z0-9]", "", x.lower())),
                    filter(
                        None,
                        [
                            self.ticker,
                            self.refinitiv_identifier_code,
                            self.refinitiv_mnemonic_code,
                            self.valoren,
                            self.sedol,
                        ],
                    ),
                )
            )
            self.search_vector = SearchVector(*names, weight="D")
            if identifiers:
                self.search_vector += SearchVector(*identifiers, weight="A")
            if isins:
                self.search_vector += SearchVector(*isins, weight="A")
            self.trigram_search_vector = f"{self.name} {' '.join(self.alternative_names)}".strip()

    def clean(self):
        if self.is_investable_universe and self.id and self.children.exists():
            raise ValidationError("An instrument in the investable universe cannot have children")
        return self

    def pre_save(self):  # noqa: C901
        if self.instrument_type:
            self.is_security = self.instrument_type.is_security
        # if self.delisted_date:
        #     self.is_security = False
        if not self.name_repr:
            self.name_repr = self.name
        if not self.founded_year and self.inception_date:
            self.founded_year = self.inception_date.year
        if not self.inception_date:
            self.inception_date = date.today() - timedelta(days=1)
        if not self.founded_year:
            self.founded_year = self.inception_date.year
        if self.level is None:
            self.level = 0
            self.rght = 0
            self.lft = 0
            self.tree_id = 0

        # ensure all identifiers are stored in uppercase and do not contain any whitespace
        for identifier_key in ["isin", "refinitiv_mnemonic_code", "ticker", "sedol", "cusip", "identifier"]:
            if identifier := getattr(self, identifier_key, None):
                setattr(self, identifier_key, identifier.upper().replace(" ", ""))
        if self.refinitiv_identifier_code:
            self.refinitiv_identifier_code = self.refinitiv_identifier_code.replace(
                " ", ""
            )  # RIC cannot be uppercased because its symbology implies meaning for lowercase characters
        self.update_search_vectors()
        if self.is_primary and (parent := self.parent):
            # we have a unique constraint on parent. We take the time to make sure no other children is already primary = True (otherwise, update will fail)
            Instrument.objects.filter(parent=parent, is_primary=True).exclude(
                source=self.source, source_id=self.source_id
            ).update(is_primary=False)
        if not self.parent:
            self.is_primary = True
        if self.id and (not self.instrument_type or not self.currency) and self.children.count() == 1:
            child = self.children.first()
            if not self.instrument_type:
                self.instrument_type = child.instrument_type
            if not self.currency:
                self.currency = child.currency

    def save(self, *args, **kwargs):
        self.pre_save()
        if self.is_primary is None:
            self.is_primary = not self.parent or (self.exchange is not None and self.parent.exchange == self.exchange)
        super().save(*args, **kwargs)

    def get_compute_str(self):
        repr = self.name_repr or self.name or ""
        repr = repr.title()  # we follow bloomberg instrument representation format
        if self.instrument_type and self.is_security:
            repr += f" {self.instrument_type.short_name}"
        if self.is_security or not self.level == 0:
            if self.identifier_repr:
                repr += f" - {self.identifier_repr}"
            # if the object has an exchange and is not a security nor a company (a quote then), we append the exchange representation
            if self.exchange and self.parent is not None and not self.is_security:
                repr += f" ({str(self.exchange)})"
        return repr

    def compute_str(self):
        return self.get_compute_str()

    def is_active_at_date(self, today: date) -> bool:
        return (
            self.inception_date is not None
            and self.inception_date <= today
            and (not self.delisted_date or self.delisted_date > today)
        )

    def update_last_valuation_date(self):
        if self.valuations.exists():
            earliest_valuation = self.valuations.earliest("date")
            last_valuation = self.valuations.latest("date").date
            if not self.inception_date or (
                not earliest_valuation.calculated and self.inception_date > earliest_valuation.date
            ):
                self.inception_date = earliest_valuation.date
            if not self.last_valuation_date or last_valuation >= self.last_valuation_date:
                self.last_valuation_date = last_valuation
        if self.prices.exists():
            last_price = self.prices.latest("date").date
            if not self.last_price_date or last_price >= self.last_price_date:
                self.last_price_date = last_price
            self.save()
            compute_metrics(last_price, basket=self)

    def get_prices(self, only_instrument_price: bool = False, **kwargs) -> Iterator[dict[str, any]]:
        qs = Instrument.objects.filter(id=self.id)
        if "market_data" in self.dl_parameters and not only_instrument_price:
            return qs.dl.market_data(**kwargs)
        # if not dataloader is found, we default to the internal instrument price dataloader
        return MarketDataDataloader(qs).market_data(**kwargs)

    def get_classifable_ancestor(self, include_self: bool = True) -> Self:
        root = self.get_root()
        if root.instrument_type and root.instrument_type.is_classifiable:
            return root

    def get_security_ancestor(self, include_self: bool = True) -> Self:
        if include_self:
            parent = self
        else:
            parent = self.parent
        while parent:
            if parent.instrument_type and parent.instrument_type.is_security:
                return parent
            parent = parent.parent

    def merge(self, merged_instrument: SelfInstrument, dispatch: bool = True, override_fields_to_copy: bool = False):
        """
        This method handle the deletion of an instrument ("merged_instrument") in favor of another one (self).

        After handling all the instrument base logic reassignment, it calls a signal that all modules can implement in
        order to implement their own reassignment logic.

        The function is atomic, it either succeed or fail (i.e. If merged_instrument is deleted, the merge was succeesful)

        Args:
            merged_instrument: The Instrument that is supposed to be merged and deleted
        """
        if self == merged_instrument:
            return
        with transaction.atomic():  # We want this to either succeed fully or fail
            # Get the base
            if dispatch:
                pre_merge.send(
                    sender=Instrument, merged_object=merged_instrument, main_object=self
                )  # default signal dispatch for the Instrument class

                # if the self type is different than Instrument, it's a polymorphic call. We fire also the pre_merge signal with this child type
                if type(self) is not Instrument:  # noqa
                    pre_merge.send(sender=self.__class__, merged_object=merged_instrument, main_object=self)

                # We refresh the reference in case the underlying signal receivers modify these objects
                self.refresh_from_db()
                merged_instrument.refresh_from_db()

            # We delete finally the merged instrument. All unlikage should have been done in the signal receivers function
            merged_instrument.delete()

            # Finally, we copy the potentially missing field from merged instrument to self (Only if self.field is None)
            field_to_copy = [
                "founded_year",
                "inception_date",
                "delisted_date",
                "name",
                "name_repr",
                "description",
                "ticker",
                "country",
                "headquarter_address",
                "headquarter_city",
                "primary_url",
                "identifier",
                "currency",
                "isin",
                "sedol",
                "valoren",
                "refinitiv_ticker",
                "refinitiv_identifier_code",
                "refinitiv_mnemonic_code",
                "exchange",
                "source",
                "source_id",
                "employees",
                "last_annual_report",
                "last_interim_report",
                "next_annual_report",
                "next_interim_report",
                "parent",
            ]
            many_fields = [
                "alternative_names",
                "old_isins",
                "additional_urls",
            ]
            for field in field_to_copy:
                if (new_value := getattr(merged_instrument, field, None)) is not None and (
                    getattr(self, field, None) is None or override_fields_to_copy
                ):
                    setattr(self, field, new_value)
            for field in many_fields:
                current_values = getattr(self, field, [])
                new_values = getattr(merged_instrument, field, [])
                setattr(self, field, list(set(current_values + new_values)))
            self.save()

    def handle_backend_lookup(self, attribute: str, method: str):
        try:
            backend = self.lookups.get(**{attribute: True}).load_backend()
            return getattr(backend, method)()
        except ObjectDoesNotExist:
            return self.__class__.objects.none()

    def technical_analysis(self, from_date: date | None = None, to_date: date | None = None):
        return TechnicalAnalysis.init_full_from_instrument(self, from_date, to_date)

    def technical_benchmark_analysis(self, from_date: date | None = None, to_date: date | None = None):
        return TechnicalAnalysis.init_close_from_instrument(self, from_date, to_date)

    def import_prices(self, start: date | None = None, end: date | None = None, **kwargs):
        if not self.is_leaf_node():
            raise ValueError("Cannot import price on a non-leaf node")
        if not start:
            start = (
                self.inception_date
                if self.inception_date
                else global_preferences_registry.manager()["wbfdm__default_start_date_historical_import"]
            )
        if not end:
            end = (
                date.today() - BDay(1)
            ).date()  # we don't import today price in case the dataloader returns duplicates (e.g. DSWS)

        # we detect when was the last date price imported before start and switch the start date from there
        with suppress(ObjectDoesNotExist):
            start = self.prices.filter(date__lte=start).latest("date").date
        # Import instrument prices
        objs = list(self.__class__.objects.filter(id=self.id).get_instrument_prices_from_market_data(start, end))
        self.bulk_save_instrument_prices(objs)
        # compute daily statistics & performances
        self.update_last_valuation_date()
        instrument_price_imported.send(sender=Instrument, instrument=self, start=start, end=end)

    @classmethod
    def parse_content_for_identifiers(cls, content: str) -> Generator[dict[str, Any], None, None]:
        for ric in re_ric(content):
            yield {"refinitiv_identifier_code": ric}
        for isin in re_isin(content):
            yield {"isin": isin}
        for ticker in re_bloomberg(content):
            yield {"ticker": ticker}
        for mnemonic in re_mnemonic(content):
            yield {"refinitiv_mnemonic_code": mnemonic}

    @classmethod
    def get_endpoint_basename(cls):
        return "wbfdm:instrument"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbfdm:instrumentrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{computed_str}}"


@shared_task(queue="portfolio")
def import_prices_as_task(instrument_id, **kwargs):
    instrument = Instrument.objects.get(id=instrument_id)
    instrument.import_prices(**kwargs)


@receiver(pre_delete, sender="wbfdm.Instrument")
def pre_delete_instrument(sender, instance, **kwargs):
    ImportedObjectProviderRelationship.objects.filter(
        content_type__in=get_ancestors_content_type(ContentType.objects.get_for_model(instance)), object_id=instance.id
    ).delete()


@receiver(pre_save, sender="wbfdm.Instrument")
def pre_save_instrument(sender, instance, raw, **kwargs):
    if not raw:
        pre_instance = None
        if instance.id:
            pre_instance = sender.objects.get(id=instance.id)
        # Remove duplicates if existings
        instance.old_isins = list(set(instance.old_isins))
        if pre_instance:
            if (
                pre_instance.isin
                and instance.isin
                and pre_instance.isin != instance.isin
                and pre_instance.isin not in instance.old_isins
            ):
                instance.old_isins = [*instance.old_isins, pre_instance.isin]
            if pre_instance.name_repr != instance.name_repr:
                # if a family member get is name representation updated, we update it for the whole family
                pre_instance.get_family().update(name_repr=instance.name_repr)

        # the instrument was manually included into the investable universe, in that case, we need to fetch the price
        if (
            instance.is_leaf_node()
            and (not pre_instance or not pre_instance.is_investable_universe)
            and instance.is_investable_universe
        ):
            import_prices_as_task.apply_async(
                (instance.id,), {"clear": True}, countdown=15
            )  # we need to introduce a countdown to avoid racing condition where save resumed after shared task picked up its instance reference.


@receiver(post_save, sender="wbfdm.Classification")
def ensure_classification_0_height(sender, instance, created, raw, **kwargs):
    # Ensure that if a leaf classification becomes non-leaf, then all instruments linked to it are updated automatically
    # with the new leaf classiciation
    if not raw and instance.parent and instance.height == 0:
        for instrument in Instrument.objects.filter(classifications=instance.parent):
            instrument.classifications.remove(instance.parent)
            instrument.classifications.add(instance)


@receiver(post_delete, sender="wbfdm.InstrumentPrice")
def post_delete_valuation(sender, instance, **kwargs):
    if not instance.calculated and instance.instrument.last_valuation_date == instance.date:
        instance.instrument.update_last_valuation_date()


class CashManager(InstrumentManager):
    def get_queryset(self) -> InstrumentQuerySet:
        return super().get_queryset().filter(instrument_type=InstrumentType.CASH)


class Cash(Instrument):
    objects = CashManager()

    def save(self, *args, **kwargs):
        self.is_cash = True
        self.dl_parameters["market_data"] = {
            "path": "wbfdm.contrib.internal.dataloaders.market_data.MarketDataDataloader"
        }

        super().save(*args, **kwargs)

    class Meta:
        proxy = True


@receiver(post_save, sender="currency.Currency")
def create_cash_from_currency(sender, instance, created, raw, **kwargs):
    if created:
        Cash.objects.get_or_create(
            instrument_type=InstrumentType.CASH, currency=instance, defaults={"name": f"Cash {instance.key}"}
        )


class EquityManager(InstrumentManager):
    def get_queryset(self) -> InstrumentQuerySet:
        return super().get_queryset().filter(instrument_type__key="equity")


class Equity(Instrument):
    objects = EquityManager()

    class Meta:
        proxy = True


@receiver(create_news_relationships, sender="wbnews.News")
def get_news_relationships_for_instruments_task(sender: type, instance: "News", **kwargs) -> shared_task:
    return run_company_extraction_llm.s(instance.title, instance.description, instance.summary)


@shared_task(queue="pms")
def detect_and_correct_financial_timeseries(
    max_days_interval: int | None = None,
    check_date: date | None = None,
    with_pelt: bool = False,
    detect_only: bool = False,
    full_reimport: bool = False,
    debug: bool = False,
):
    """Detects and corrects anomalies in financial time series data for instruments.

    Analyzes price data using statistical methods to identify outliers and change points,
    then triggers price reimport for affected date ranges when corrections are needed.

    Args:
        max_days_interval: Maximum lookback window in days for analysis (None = all history)
        check_date: Reference date for analysis (defaults to current date)
        with_pelt: Enable Pelt's change point detection alongside basic z-score outlier detection
        detect_only: Run detection without performing data correction/reimport
        full_reimport: Reimport entire price history when corruption detected (requires max_days_interval=None)
        debug: Show progress bar during instrument processing

    """
    if not check_date:
        check_date = date.today()
    gen = (
        Instrument.investable_universe.filter(is_managed=False)
        .filter_active_at_date(check_date)
        .exclude(source="dsws")
    )
    if debug:
        gen = tqdm(gen, total=gen.count())
    for instrument in gen:
        prices = instrument.valuations.all()
        if max_days_interval:
            prices = prices.filter(date__gte=check_date - timedelta(days=max_days_interval))
        # construct the price timeseries
        prices_series = (
            pd.DataFrame(
                prices.filter_only_valid_prices().values_list("date", "net_value"), columns=["date", "net_value"]
            )
            .set_index("date")["net_value"]
            .astype(float)
            .sort_index()
        )
        if not prices_series.empty:
            outliers = outlier_detection(prices_series).index.tolist()
            # if pelt enable, add the outliers found by the PELT model
            if with_pelt:
                outliers.extend(statistical_change_point_detection(prices_series).index.tolist())
            if outliers:
                logger.info(f"Abnormal change point detected for {instrument} at {outliers}.")
                if not detect_only:
                    # for a full reimport, we delete the whole existing price series and reimport since inception
                    if full_reimport and not max_days_interval:
                        start_import_date = instrument.inception_date
                        end_import_date = check_date
                        instrument.prices.filter(assets__isnull=True).delete()
                    else:
                        start_import_date = min(outliers) - timedelta(days=7)
                        end_import_date = max(outliers) + timedelta(days=7)
                    logger.info(f"Reimporting price from {start_import_date} to {end_import_date}...")
                    instrument.import_prices(start=start_import_date, end=end_import_date)


@receiver(investable_universe_updated, sender="wbfdm.Instrument")
def investable_universe_change_point_detection(*args, end_date: date | None = None, **kwargs):
    detect_and_correct_financial_timeseries.delay(check_date=end_date, max_days_interval=365)



from datetime import date as date_lib
from typing import Self, Type

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models.fields.json import KeyTextTransform
from django.db.models.functions import Cast, NullIf
from wbcore.models import WBModel

from .backends.base import AbstractBackend
from .dto import Metric, MetricKey
from .registry import backend_registry


class InstrumentMetric(models.Model):
    basket_content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, related_name="instrument_metrics")
    basket_id = models.PositiveIntegerField()
    basket_repr = models.CharField(max_length=256)

    basket = GenericForeignKey("basket_content_type", "basket_id")

    instrument = models.ForeignKey(
        "wbfdm.Instrument",
        on_delete=models.CASCADE,
        related_name="metrics",
        null=True,
        blank=True,
        help_text="Instrument where this metric belongs to. If null, the metric is applicable to all instruments related to the basket",
    )
    date = models.DateField(
        verbose_name="Metric Date", null=True, blank=True, help_text="If date is null, the metric is considered static"
    )
    key = models.CharField(max_length=255, verbose_name="Metric Key")
    metrics = models.JSONField(default=dict, verbose_name="Metrics", encoder=DjangoJSONEncoder)

    parent_metric = models.ForeignKey(
        "self",
        related_name="dependent_metrics",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        verbose_name="Parent Metric",
    )

    def save(self, *args, **kwargs):
        self.basket_repr = str(self.basket)
        super().save(*args, **kwargs)

    def __str__(self) -> str:
        repr = f"{self.basket} - {self.key}"
        if self.date:
            repr += f"({self.date})"
        return repr

    @classmethod
    def update_or_create_from_metric(cls, metric: Metric, parent_instrument_metric: Self | None = None):
        """
        Update or create an InstrumentMetric instance based on a given Metric DTO object.

        Args:
            metric (Metric): The DTO to base the creation or update on.
            parent_instrument_metric (Optional[InstrumentMetric]): The parent of the created/updated InstrumentMetric instance, if any. Defaults to None.

        Returns:
            None

        Side Effects:
            - Creates or updates an InstrumentMetric instance in the database.
            - Recursively creates or updates dependent InstrumentMetric instances.
        """
        base_metric, _ = InstrumentMetric.objects.update_or_create(
            basket_id=metric.basket_id,
            basket_content_type_id=metric.basket_content_type_id,
            instrument_id=metric.instrument_id,
            date=metric.date,
            key=metric.key,
            defaults={
                "metrics": metric.metrics,
                "parent_metric": parent_instrument_metric,
            },
        )
        for dependency_metric in metric.dependency_metrics:
            cls.update_or_create_from_metric(dependency_metric, parent_instrument_metric=base_metric)

    @classmethod
    def annotate_with_metrics(
        cls,
        queryset: models.QuerySet,
        metric_key: MetricKey,
        metric_basket_class: Type[WBModel],
        val_date: date_lib | None = None,
        basket_label: str = "id",
        instrument_label: str | None = None,
    ) -> models.QuerySet:
        """
        Annotate metrics to a queryset related to the given metric key and basket class.

        Args:
            queryset (models.QuerySet): The queryset to be annotated.
            metric_key (MetricKey): The metric key or MetricKey instance to use for annotation.
            metric_basket_class (Type[WBModel]): The basket class associated with the metric.
            val_date (date, optional): The date for the metrics. Defaults to None.
            basket_label (str, optional): The label to identify the basket in the subquery queryset. Defaults to "id".
            instrument_label (Union[str, None], optional): The label to identify the instrument in the subquery queryset, if any. Defaults to None.

        Returns:
            models.QuerySet: The annotated queryset with the metrics included. (with field keys as returned by MetricField.get_fields)

        Side Effects:
            - Annotates the given queryset with subqueries for the specified metrics.
        """

        backend_class: AbstractBackend = backend_registry[metric_key, metric_basket_class]

        content_type = ContentType.objects.get_for_model(backend_class.BASKET_MODEL_CLASS)
        subquery = InstrumentMetric.objects.filter(
            basket_content_type=content_type,
            basket_id=models.OuterRef(basket_label),
            key=metric_key.key,
            date=val_date,
        )
        if instrument_label:
            subquery = subquery.filter(instrument=models.OuterRef(instrument_label))
        queryset = queryset.annotate(
            **{
                k: models.Subquery(
                    subquery.annotate(
                        casted_metric=Cast(
                            NullIf(models.F("metrics__" + v), models.Value("null")),
                            output_field=models.FloatField(),
                        )
                    ).values("casted_metric")[:1]
                )
                for k, v in metric_key.subfields_filter_map.items()
            }
        )

        # annotate extra subfields
        queryset = queryset.annotate(
            **{
                f"{metric_key.key}_{extra_subfield.key}": models.Subquery(
                    subquery.annotate(
                        casted_metric=Cast(
                            KeyTextTransform(extra_subfield.key, "metrics"), output_field=extra_subfield.field_type()
                        ),
                    ).values("casted_metric")[:1]
                )
                for extra_subfield in metric_key.extra_subfields
            }
        )

        return queryset

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "metric:instrumentmetric"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "metric:instrumentmetricrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{key}} - {{date}} - {{basket_repr}}"

    class Meta:
        verbose_name = "Instrument Metric"
        verbose_name_plural = "Instrument Metrics"
        constraints = [
            models.UniqueConstraint(
                name="unique_instrument_metric",
                fields=["basket_content_type", "basket_id", "instrument", "date", "key"],
                nulls_distinct=False,
            )
        ]
        indexes = [
            models.Index(
                fields=["basket_content_type", "basket_id", "key", "date"],
            )
        ]



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

T = TypeVar("T")


class Sync(Generic[T], ABC):
    @abstractmethod
    def update(self):
        """Updates all items"""
        ...

    @abstractmethod
    def update_or_create_item(self, external_id: int) -> T:
        """Updates or creates a single item from an external id and returns it"""
        ...

    @abstractmethod
    def update_item(self, item: T) -> T:
        """Updates a single item and returns it"""
        ...

    @abstractmethod
    def get_item(self, external_id: int) -> dict:
        """Retrieves a dictionairy representation of the item"""
        ...

    @abstractmethod
    def trigger_partial_update(self):
        """Figures out if an update is necessary and then tries to update the entire database in the most efficiant way"""
        ...



from wbcore.metadata.configs.titles import TitleViewConfig


class ExchangeTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Exchanges"

    def get_instance_title(self):
        return "Exchange"

    def get_create_title(self):
        return "New Exchange"



from datetime import date

import pandas as pd
import pytest
from faker import Faker
from wbcore.models import DynamicDecimalField, DynamicFloatField

from wbfdm.models import Instrument, InstrumentPrice, RelatedInstrumentThroughModel

fake = Faker()


@pytest.mark.django_db
class TestInstrumentPriceModel:
    def test_init(self, instrument_price):
        assert instrument_price.id is not None

    def test_str(self, instrument_price):
        assert (
            str(instrument_price)
            == f"{instrument_price.instrument.name}: {instrument_price.net_value} {instrument_price.date:%d.%m.%Y}"
        )

    def test_previous_price(self, instrument, instrument_price_factory):
        previous_price = instrument_price_factory.create(instrument=instrument, date=date(2009, 12, 31))
        price = instrument_price_factory.create(instrument=instrument, date=date(2010, 1, 1))
        assert price.previous_price == previous_price

    def test_previous_price_does_not_exist(self, instrument_price):
        assert instrument_price.previous_price is None

    def test_next_price_price(self, instrument, instrument_price_factory):
        next_price = instrument_price_factory.create(instrument=instrument, date=date(2010, 1, 4))
        price = instrument_price_factory.create(instrument=instrument, date=date(2010, 1, 1))
        assert price.next_price == next_price

    def test_next_price_does_not_exist(self, instrument_price):
        assert instrument_price.next_price is None

    def test_subquery_closest_value(self, instrument, instrument_price_factory):
        instrument_price_factory.create_batch(5, instrument=instrument)
        latest_valid_price = InstrumentPrice.objects.latest("date")
        queryset = Instrument.objects.all().annotate(
            net_value=InstrumentPrice.subquery_closest_value(
                "net_value", val_date=latest_valid_price.date, instrument_pk_name="pk"
            )
        )

        assert queryset.get(id=instrument.id).net_value == latest_valid_price.net_value

    def test_compute_and_update_statistics(self, weekday, instrument_factory, instrument_price_factory):
        product = instrument_factory.create()
        benchmark = instrument_factory.create()
        RelatedInstrumentThroughModel.objects.create(
            instrument=product,
            related_instrument=benchmark,
            related_type=RelatedInstrumentThroughModel.RelatedTypeChoices.BENCHMARK,
        )
        risk_free_rate = instrument_factory.create()
        RelatedInstrumentThroughModel.objects.create(
            instrument=product,
            related_instrument=risk_free_rate,
            related_type=RelatedInstrumentThroughModel.RelatedTypeChoices.RISK_INSTRUMENT,
        )

        for _d in pd.date_range(end=weekday, periods=5, freq="B"):
            instrument_price_factory.create(
                calculated=False, instrument=benchmark, date=_d.date(), sharpe_ratio=None, correlation=None, beta=None
            )
            instrument_price_factory.create(
                calculated=False,
                instrument=risk_free_rate,
                date=_d.date(),
                sharpe_ratio=None,
                correlation=None,
                beta=None,
            )
            instrument_price_factory.create(
                calculated=False, instrument=product, date=_d.date(), sharpe_ratio=None, correlation=None, beta=None
            )

        p = product.valuations.get(date=weekday)
        p.compute_and_update_statistics(min_period=5)
        assert p.sharpe_ratio
        assert p.correlation
        assert p.beta

    # Compute functions tests
    #
    # @pytest.mark.parametrize("instrument_price__custom_beta_180d", [None])
    # def test_compute_custom_beta_180d(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_custom_beta_180d")
    #     assert isinstance(instrument_price._meta.get_field("custom_beta_180d"), DynamicFloatField)
    #
    # @pytest.mark.parametrize("instrument_price__custom_beta_1y", [None])
    # def test_compute_custom_beta_1y(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_custom_beta_1y")
    #     assert isinstance(instrument_price._meta.get_field("custom_beta_1y"), DynamicFloatField)
    #
    # @pytest.mark.parametrize("instrument_price__custom_beta_2y", [None])
    # def test_compute_custom_beta_2y(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_custom_beta_2y")
    #     assert isinstance(instrument_price._meta.get_field("custom_beta_2y"), DynamicFloatField)
    #
    # @pytest.mark.parametrize("instrument_price__custom_beta_3y", [None])
    # def test_compute_custom_beta_3y(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_custom_beta_3y")
    #     assert isinstance(instrument_price._meta.get_field("custom_beta_3y"), DynamicFloatField)
    #
    # @pytest.mark.parametrize("instrument_price__custom_beta_5y", [None])
    # def test_compute_custom_beta_5y(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_custom_beta_5y")
    #     assert isinstance(instrument_price._meta.get_field("custom_beta_5y"), DynamicFloatField)
    #
    # @pytest.mark.parametrize(
    #     "instrument_price__free_float_ratio, instrument_price__outstanding_shares", [(None, decimal.Decimal(10))]
    # )
    # def test_compute_free_float_ratio(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_free_float_ratio")
    #     assert isinstance(instrument_price._meta.get_field("free_float_ratio"), DynamicFloatField)
    #     assert instrument_price.free_float_ratio == instrument_price.free_float / float(
    #         instrument_price.outstanding_shares
    #     )
    #
    # @pytest.mark.parametrize(
    #     "instrument_price__short_interest_ratio, instrument_price__outstanding_shares", [(None, decimal.Decimal(10))]
    # )
    # def test_compute_short_interest_ratio(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_short_interest_ratio")
    #     assert isinstance(instrument_price._meta.get_field("short_interest_ratio"), DynamicFloatField)
    #     assert instrument_price.short_interest_ratio == instrument_price.short_interest / float(
    #         instrument_price.outstanding_shares
    #     )
    #
    # @pytest.mark.parametrize("instrument_price__net_value_usd", [None])
    # def test_compute_net_value_usd(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_net_value_usd")
    #     assert isinstance(instrument_price._meta.get_field("net_value_usd"), DynamicFloatField)
    #     assert instrument_price.net_value_usd == instrument_price.currency_fx_usd * float(instrument_price.net_value)
    #
    # @pytest.mark.parametrize("instrument_price__volume_usd", [None])
    # def test_compute_volume_usd(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_volume_usd")
    #     assert isinstance(instrument_price._meta.get_field("volume_usd"), DynamicFloatField)
    #     assert instrument_price.volume_usd == instrument_price.currency_fx_usd * instrument_price.volume * float(
    #         instrument_price.net_value
    #     )
    #
    # @pytest.mark.parametrize("instrument_price__volume_50d_usd", [None])
    # def test_compute_volume_50d_usd(self, instrument_price):
    #     assert hasattr(instrument_price, "_compute_volume_50d_usd")
    #     assert isinstance(instrument_price._meta.get_field("volume_50d_usd"), DynamicFloatField)
    #     assert instrument_price.volume_50d_usd == instrument_price.net_value_usd * instrument_price.volume_50d
    #
    # @pytest.mark.parametrize("instrument_price__currency_fx_usd", [None])
    # def test_compute_currency_fx_usd(self, instrument_price):
    #     fx_rate = instrument_price.instrument.currency.fx_rates.get(date=instrument_price.date)
    #     instrument_price.save(update_all_dynamic_fields=True)
    #     assert hasattr(instrument_price, "_compute_currency_fx_usd")
    #     assert isinstance(instrument_price._meta.get_field("currency_fx_usd"), DynamicFloatField)
    #     assert instrument_price.currency_fx_usd == 1 / float(fx_rate.value)

    @pytest.mark.parametrize("instrument_price__gross_value", [None])
    def test_compute_gross_value(self, instrument_price):
        assert hasattr(instrument_price, "_compute_gross_value")
        assert isinstance(instrument_price._meta.get_field("gross_value"), DynamicDecimalField)
        assert instrument_price.gross_value == instrument_price.net_value

    @pytest.mark.parametrize("instrument_price__volume_50d", [None])
    def test_compute_volume_50d(self, instrument_price, instrument_price_factory):
        assert hasattr(instrument_price, "_compute_volume_50d")
        assert isinstance(instrument_price._meta.get_field("volume_50d"), DynamicFloatField)

        sum_volume_50d = instrument_price.volume
        for _d in pd.date_range(end=instrument_price.date, periods=50, freq="B", inclusive="left"):
            p = instrument_price_factory.create(date=_d, instrument=instrument_price.instrument)
            sum_volume_50d += p.volume
        instrument_price.save(update_all_dynamic_fields=True)
        assert instrument_price.volume_50d == sum_volume_50d / 50

    # @pytest.mark.parametrize("instrument_price__volume_200d", [None])
    # def test_compute_volume_200d(self, instrument_price, instrument_price_factory):
    #     assert hasattr(instrument_price, "_compute_volume_200d")
    #     assert isinstance(instrument_price._meta.get_field("volume_200d"), DynamicFloatField)
    #
    #     sum_volume_200d = instrument_price.volume
    #     for _d in pd.date_range(end=instrument_price.date, periods=200, freq="B", inclusive="left"):
    #         p = instrument_price_factory.create(date=_d, instrument=instrument_price.instrument)
    #         sum_volume_200d += p.volume
    #     instrument_price.save(update_all_dynamic_fields=True)
    #     assert instrument_price.volume_200d == sum_volume_200d / 200

    # @pytest.mark.parametrize("instrument_price__market_capitalization_usd", [None])
    # def test_compute_market_capitalization_usd(self, instrument_price):
    #     assert hasattr(instrument_price, "market_capitalization_usd")
    #     assert isinstance(instrument_price._meta.get_field("market_capitalization_usd"), DynamicFloatField)
    #     assert (
    #         instrument_price.market_capitalization_usd
    #         == instrument_price.market_capitalization * instrument_price.currency_fx_usd
    #     )

    # def test_compute_performance_1d(self, instrument_price_factory):
    #     previous_price = instrument_price_factory.create()
    #     instrument_price = instrument_price_factory.create(
    #         instrument=previous_price.instrument,
    #         calculated=previous_price.calculated,
    #         date=previous_price.date + BDay(1),
    #         performance_1d=None,
    #     )
    #     assert hasattr(instrument_price, "performance_1d")
    #     assert isinstance(instrument_price._meta.get_field("performance_1d"), DynamicDecimalField)
    #     assert instrument_price.performance_1d == instrument_price.net_value / previous_price.net_value - 1
    #
    # def test_compute_performance_7d(self, instrument_price_factory):
    #     previous_price = instrument_price_factory.create()
    #     instrument_price = instrument_price_factory.create(
    #         instrument=previous_price.instrument,
    #         calculated=previous_price.calculated,
    #         date=previous_price.date + BDay(7),
    #         performance_7d=None,
    #     )
    #     assert hasattr(instrument_price, "performance_7d")
    #     assert isinstance(instrument_price._meta.get_field("performance_7d"), DynamicDecimalField)
    #     assert instrument_price.performance_7d == instrument_price.net_value / previous_price.net_value - 1
    #
    # def test_compute_performance_30d(self, instrument_price_factory):
    #     previous_price = instrument_price_factory.create()
    #     instrument_price = instrument_price_factory.create(
    #         instrument=previous_price.instrument,
    #         calculated=previous_price.calculated,
    #         date=previous_price.date + BDay(30),
    #         performance_30d=None,
    #     )
    #     assert hasattr(instrument_price, "performance_30d")
    #     assert isinstance(instrument_price._meta.get_field("performance_30d"), DynamicDecimalField)
    #     assert instrument_price.performance_30d == instrument_price.net_value / previous_price.net_value - 1
    #
    # def test_compute_performance_90d(self, instrument_price_factory):
    #     previous_price = instrument_price_factory.create()
    #     instrument_price = instrument_price_factory.create(
    #         instrument=previous_price.instrument,
    #         calculated=previous_price.calculated,
    #         date=previous_price.date + BDay(90),
    #         performance_90d=None,
    #     )
    #     assert hasattr(instrument_price, "performance_90d")
    #     assert isinstance(instrument_price._meta.get_field("performance_90d"), DynamicDecimalField)
    #     assert instrument_price.performance_90d == instrument_price.net_value / previous_price.net_value - 1
    #
    # def test_compute_performance_365d(self, instrument_price_factory):
    #     previous_price = instrument_price_factory.create()
    #     instrument_price = instrument_price_factory.create(
    #         instrument=previous_price.instrument,
    #         calculated=previous_price.calculated,
    #         date=previous_price.date + BDay(365),
    #         performance_365d=None,
    #     )
    #     assert hasattr(instrument_price, "performance_365d")
    #     assert isinstance(instrument_price._meta.get_field("performance_365d"), DynamicDecimalField)
    #     assert instrument_price.performance_365d == instrument_price.net_value / previous_price.net_value - 1
    #
    # @pytest.mark.parametrize("instrument_price__performance_ytd", [None])
    # def test_compute_performance_ytd(self, instrument_price, instrument_price_factory):
    #     previous_price = instrument_price_factory.create(
    #         instrument=instrument_price.instrument,
    #         calculated=instrument_price.calculated,
    #         date=(date(instrument_price.date.year, 1, 1) - BDay(0)).date(),
    #         performance_ytd=None,
    #     )
    #     instrument_price.save(update_all_dynamic_fields=True)
    #     assert hasattr(instrument_price, "performance_ytd")
    #     assert isinstance(instrument_price._meta.get_field("performance_ytd"), DynamicDecimalField)
    #     assert instrument_price.performance_ytd == instrument_price.net_value / previous_price.net_value - 1
    #
    # def test_compute_performance_inception(self, instrument_price_factory):
    #     inception_price = instrument_price_factory.create()
    #     inception_price.instrument.inception_date = inception_price.date
    #     inception_price.instrument.save()
    #     instrument_price_factory.create(
    #         instrument=inception_price.instrument,
    #         calculated=inception_price.calculated,
    #         date=inception_price.date + BDay(1),
    #         performance_inception=None,
    #     )
    #     instrument_price = instrument_price_factory.create(
    #         instrument=inception_price.instrument,
    #         calculated=inception_price.calculated,
    #         date=inception_price.date + BDay(2),
    #         performance_inception=None,
    #     )
    #     assert hasattr(instrument_price, "performance_inception")
    #     assert isinstance(instrument_price._meta.get_field("performance_inception"), DynamicDecimalField)
    #     assert instrument_price.performance_inception == instrument_price.net_value / inception_price.net_value - 1



from django.contrib.postgres.fields import ArrayField
from django.db import models
from wbcore.contrib.io.mixins import ImportMixin

from wbfdm.import_export.handlers.private_equities import DealImportHandler


class Deal(ImportMixin, models.Model):
    import_export_handler_class = DealImportHandler

    class Types(models.TextChoices):
        DEAL = "DEAL", "Deal"
        FUNDING = "FUNDING", "Funding"
        INVESTMENT = "INVESTMENT", "Investment"
        PORTFOLIO_EXIT = "PORTFOLIO_EXIT", "Portfolio Exit"

    type = models.CharField(
        default=Types.DEAL, choices=Types.choices, max_length=14, verbose_name="Type", help_text="The deal type"
    )
    external_id = models.CharField(max_length=64, blank=True, null=True)
    date = models.DateField()
    equity = models.ForeignKey(
        "wbfdm.Instrument",
        related_name="deals",
        on_delete=models.CASCADE,
        limit_choices_to=models.Q(instrument_type__key="equity"),
    )
    transaction_amount = models.FloatField(help_text="Deal Size (in millions")
    investors = models.ManyToManyField(
        "wbfdm.Instrument",
        related_name="invested_deals",
        blank=True,
        verbose_name="Investors",
        help_text="Investors",
    )

    funding_round = models.CharField(max_length=128, verbose_name="Funding Round")
    funding_round_category = models.CharField(max_length=128, verbose_name="Funding Round Category")

    valuation = models.FloatField(
        verbose_name="Valuaton",
        blank=True,
        null=True,
        help_text="Valuation of the funded organization after this transaction (in Millions USD).",
    )
    valuation_estimated = models.BooleanField(
        default=False, verbose_name="Is valuation estimated", help_text="True if the valuation is an estimate"
    )
    valuation_source_type = models.CharField(
        max_length=24,
        blank=True,
        null=True,
        help_text="The source type of the valuation",
    )
    valuation_media_mention_source_urls = ArrayField(
        models.URLField(),
        blank=True,
        null=True,
        help_text="List of URLs used to source the valuation for the Media Mentions source type.",
    )

    def __str__(self) -> str:
        return f"{self.equity} - {self.date}"



import pandas as pd

from wbfdm.enums import ESG
from wbfdm.models.instruments.instruments import InstrumentQuerySet


def get_esg_df(instruments: InstrumentQuerySet, esg: ESG, **kwargs) -> pd.Series:
    df = pd.DataFrame(instruments.dl.esg(values=[esg]))
    if not df.empty:
        return df.pivot_table(
            index="instrument_id", values="value", columns="factor_code", aggfunc="first", dropna=False
        )[esg.value]
    else:
        return pd.Series(dtype="float64")



import numpy as np
import pandas as pd
import pytest
from faker import Faker
from wbcore.contrib.currency.factories import CurrencyFXRatesFactory
from wbcore.contrib.currency.models import CurrencyFXRates

from wbfdm.analysis.esg.enums import ESGAggregation
from wbfdm.analysis.esg.esg_analysis import DataLoader

fake = Faker()

WEIGHTS = [0.15, 0.05, 0.2, 0.1, 0.5]
ENTERPRISE_VALUE_INCLUDED_CASH = [1e6, 2e6, 3e6, 4e6, 1e6]
CURRENT_VALUE_INVESTMENT_FACTOR = [1.0, 1.0, 1.0, 1.2, 0.2]


@pytest.mark.django_db
class TestESGDataLoader:
    @pytest.fixture
    def dataloader(self, weekday):
        index = [1, 2, 3, 4, 5]

        weight = pd.Series(WEIGHTS, index=index)
        esg_data = pd.Series(np.random.randint(1, 100, size=4), index=[1, 2, 3, 4])  # we remove one data point
        total_value_fx_usd = pd.Series(np.random.randint(1, 100, size=5))  # we remove one data point
        CurrencyFXRatesFactory.create(currency__key="EUR", date=weekday)

        dataloader = DataLoader(weight, esg_data, weekday, total_value_fx_usd=total_value_fx_usd)
        dataloader.__dict__["enterprise_value_included_cash"] = pd.Series(ENTERPRISE_VALUE_INCLUDED_CASH, index=index)
        dataloader.__dict__["current_value_investment_factor"] = pd.Series(
            CURRENT_VALUE_INVESTMENT_FACTOR, index=index
        )
        dataloader.__dict__["nace_section_code"] = pd.Series(["C", "C", "J", "M", "M"], index=index)
        return dataloader

    def test_dataloader(self, dataloader):
        assert dataloader.weights_in_coverage.to_dict() == {1: 0.15, 2: 0.05, 3: 0.2, 4: 0.1}

    def test_get_percentage_sum(self, dataloader):
        dataloader.esg_data = pd.Series(["Yes", "No Evidence", "No", "No", "No"], index=[1, 2, 3, 4, 5])
        res = dataloader._get_percentage_sum("No Evidence")
        assert res.tolist() == [0.0, 0.05, 0.0, 0.0]

    def test_get_weighted_avg_normalized(self, dataloader):
        res = dataloader._get_weighted_avg_normalized()
        log = dataloader.intermediary_logs[0]

        # ensure the intermediary log data is set properly
        assert log.series.name == "weights_normalized"
        assert log.series.to_dict() == {1: 0.3, 2: 0.1, 3: 0.4, 4: 0.2}
        assert res.to_dict() == {
            1: 0.3 * dataloader.esg_data.iloc[0],
            2: 0.1 * dataloader.esg_data.iloc[1],
            3: 0.4 * dataloader.esg_data.iloc[2],
            4: 0.2 * dataloader.esg_data.iloc[3],
        }

    def test_get_weighted_avg_normalized_per_category(self, dataloader):
        res = dataloader._get_weighted_avg_normalized_per_category()
        intermediary_logs = list(filter(lambda x: not x.series.empty, dataloader.intermediary_logs))
        log_c = intermediary_logs[0]
        log_j = intermediary_logs[1]
        log_m = intermediary_logs[2]
        weights_normalized = [
            0.15 / (0.15 + 0.05),
            0.05 / (0.15 + 0.05),
            0.2 / 0.2,
            0.1 / (0.1 + 0.5),
            0.5 / (0.1 + 0.5),
        ]
        # ensure the intermediary log data is set properly
        assert log_c.series.name == "weights_normalized_sector_c"
        assert log_c.series.to_dict() == {
            1: weights_normalized[0],
            2: weights_normalized[1],
            # 3: weights_normalized[2],
            # 4: weights_normalized[3],
            #             5: weights_normalized[4],
        }
        assert log_j.series.name == "weights_normalized_sector_j"
        assert log_j.series.to_dict() == {
            3: weights_normalized[2],
        }
        assert log_m.series.name == "weights_normalized_sector_m"
        assert log_m.series.to_dict() == {
            4: weights_normalized[3],
            5: weights_normalized[4],
        }
        assert res.to_dict() == {
            1: weights_normalized[0] * dataloader.esg_data.iloc[0],
            2: weights_normalized[1] * dataloader.esg_data.iloc[1],
            3: weights_normalized[2] * dataloader.esg_data.iloc[2],
            4: weights_normalized[3] * dataloader.esg_data.iloc[3],
        }

    def test_get_investor_allocation(self, dataloader):
        exposure = pd.Series(np.random.rand(5), index=dataloader.weights.index)
        res = dataloader._get_investor_allocation(exposure)
        fx_rate = CurrencyFXRates.objects.get(currency__key="EUR", date=dataloader.val_date)
        exposure_eur = exposure * float(fx_rate.value)

        assert dataloader.intermediary_logs[0].series.to_dict() == exposure.to_dict()
        assert (
            dataloader.intermediary_logs[1].series.to_dict() == exposure_eur.to_dict()
        )  # exposure converted into eur

        exposure_with_cvi = [exposure_eur.iloc[i] * CURRENT_VALUE_INVESTMENT_FACTOR[i] for i in range(5)]
        assert dataloader.intermediary_logs[2].series.tolist() == exposure_with_cvi

        rebase_factor = sum(exposure_with_cvi[:4]) / sum(exposure_with_cvi)

        exposure_normalized = [exposure_with_cvi[i] / rebase_factor for i in range(4)]
        assert dataloader.intermediary_logs[3].series.tolist() == exposure_normalized

        attribution_factor = [exposure_normalized[i] / ENTERPRISE_VALUE_INCLUDED_CASH[i] for i in range(4)]
        assert dataloader.intermediary_logs[4].series.dropna().tolist() == attribution_factor

        assert res.to_dict() == {
            1: dataloader.esg_data.iloc[0] * attribution_factor[0],
            2: dataloader.esg_data.iloc[1] * attribution_factor[1],
            3: dataloader.esg_data.iloc[2] * attribution_factor[2],
            4: dataloader.esg_data.iloc[3] * attribution_factor[3],
        }

    def test_compute_percentage_sum(self, dataloader):
        pd.testing.assert_series_equal(
            dataloader.compute(ESGAggregation.FOSSIL_FUEL_EXPOSURE),
            dataloader._get_percentage_sum("Yes"),
            check_exact=True,
        )

    def test_compute_weighted_avg_normalized(self, dataloader):
        res = dataloader.compute(ESGAggregation.GHG_INTENSITY_OF_COMPAGNIES)
        assert res.empty is not None
        pd.testing.assert_series_equal(
            dataloader.compute(ESGAggregation.GHG_INTENSITY_OF_COMPAGNIES),
            dataloader._get_weighted_avg_normalized(),
            check_exact=True,
        )

    def test_compute_weighted_avg_normalized_per_category(self, dataloader):
        res = dataloader.compute(ESGAggregation.ENERGY_CONSUMPTION_INTENSITY_PER_SECTOR)
        assert res.empty is not None
        pd.testing.assert_series_equal(res, dataloader._get_weighted_avg_normalized_per_category(), check_exact=True)

    def test_compute_investor_allocation(self, dataloader):
        res = dataloader.compute(ESGAggregation.GHG_EMISSIONS_SCOPE_1)
        assert res.empty is not None
        pd.testing.assert_series_equal(
            res, dataloader._get_investor_allocation(dataloader.total_value_fx_usd), check_exact=True
        )

        res = dataloader.compute(ESGAggregation.GHG_EMISSIONS_SCOPE_1)
        assert res.empty is not None
        pd.testing.assert_series_equal(
            dataloader.compute(ESGAggregation.EMISSIONS_TO_WATER),
            dataloader._get_investor_allocation_per_million(dataloader.weights * 1000000),
            check_exact=True,
        )



from wbfdm.import_export.backends.refinitiv.daily_fundamental import DEFAULT_MAPPING

from .utils import parse_daily_fundamental_data


def parse(import_source):
    return parse_daily_fundamental_data(import_source, DEFAULT_MAPPING, extra_normalization_map={"free_cash": 1000})



from datetime import date, timedelta
from typing import TYPE_CHECKING, Literal

import pandas as pd
from pandas.tseries.offsets import BDay
from stockstats import StockDataFrame

from wbfdm.analysis.technical_analysis.traces import TechnicalAnalysisTraceFactory
from wbfdm.enums import MarketData

if TYPE_CHECKING:
    from wbfdm.models import Instrument


class TechnicalAnalysis:
    def __init__(self, instrument: "Instrument", sdf: StockDataFrame):
        self.instrument = instrument
        self._sdf = sdf

    def trace_factory(self) -> TechnicalAnalysisTraceFactory:
        return TechnicalAnalysisTraceFactory(self)

    @property
    def instrument_name(self):
        return self.instrument.name

    @classmethod
    def init_from_dataloader(cls, instrument: "Instrument", dataloader):
        try:
            return cls(
                instrument,
                StockDataFrame.retype(
                    pd.DataFrame(dataloader)
                    .rename(columns={"valuation_date": "date"})
                    .set_index("date")
                    .sort_index()
                    .bfill()
                ),
            )
        except KeyError:
            return cls(instrument, StockDataFrame(columns=["date", "close", "volume", "calculated"]))

    @classmethod
    def init_full_from_instrument(
        cls, instrument: "Instrument", from_date: date | None = None, to_date: date | None = None
    ):
        return cls.init_from_dataloader(
            instrument,
            instrument.__class__.objects.filter(id=instrument.id).dl.market_data(from_date=from_date, to_date=to_date),
        )

    @classmethod
    def init_close_from_instrument(
        cls, instrument: "Instrument", from_date: date | None = None, to_date: date | None = None
    ):
        return cls.init_from_dataloader(
            instrument,
            instrument.__class__.objects.filter(id=instrument.id).dl.market_data(
                values=[MarketData.CLOSE], from_date=from_date, to_date=to_date
            ),
        )

    def add_sma(self, window, field="close"):
        self._sdf[[f"{field}_{window}_sma"]]

    def add_roc(self, window, field="close"):
        self._sdf[[f"{field}_{window}_roc"]]

    def add_shift(self, window, field="close"):
        self._sdf[[f"{field}_{window}_s"]]

    def add_delta(self, window, field="close"):
        self._sdf[[f"{field}_{window}_d"]]

    def add_drawdown(self, field="close"):
        self._sdf["drawdown"] = self._sdf[field] - self._sdf[field].cummax()

    def add_macd(self):
        self._sdf[["macd", "macdh", "macds"]]

    def add_bollinger(self, window=14):
        self._sdf[[f"boll_{window}", f"boll_ub_{window}", f"boll_lb_{window}"]]
        self._sdf.bfill()

    def add_energy_index(self):
        self._sdf[["cr", "cr-ma1", "cr-ma2", "cr-ma3"]]

    def add_williams_index(self, window=14):
        self._sdf[[f"wr_{window}"]]

    def add_volume_variation_index(self, window=26):
        self._sdf[[f"vr_{window}"]]

    def add_return(self, return_type: Literal["log"] | Literal["normal"] = "normal"):
        if return_type == "log":
            self._sdf[["log-ret"]]
        elif return_type == "normal":
            self._sdf["ret"] = (self._sdf["close"] - self._sdf["close_-1_s"]) / self._sdf["close_-1_s"]

    def add_cumulative_return(self, return_type: Literal["log"] | Literal["normal"] = "normal"):
        if return_type == "log":
            self._sdf["cum-log-ret"] = (1 + self._sdf[["log-ret"]]).cumprod() - 1
        elif return_type == "normal":
            self.add_return("normal")
            self._sdf["cum-ret"] = (1 + self._sdf[["ret"]]).cumprod() - 1

    def get_performance_between_dates(self, from_date: date, to_date: date | None = None) -> float:
        if to_date is None:
            to_date = date.today()

        close = self.df[(self.df.index >= (pd.Timestamp(from_date) - BDay(1)).date()) & (self.df.index <= to_date)][
            "close"
        ]
        if close.empty:
            return 0
        return (close[-1] - close[0]) / close[0]

    def get_performance_year_to_date(self) -> float:
        return self.get_performance_between_dates(from_date=date.today().replace(month=1, day=1))

    def get_performance_months(self, months: int) -> float:
        return self.get_performance_between_dates(from_date=date.today() - timedelta(30 * months))

    def get_performances_dataframe(self, freq: Literal["Y"] | Literal["ME"]) -> pd.DataFrame:
        df = self.df
        df.index = pd.to_datetime(df.index)
        df = df.groupby(pd.Grouper(freq=freq)).last()
        df["performance"] = df["close"].pct_change()
        return df[["close", "performance"]]

    def get_annual_performances_dataframe(self):
        return self.get_performances_dataframe("Y")

    def get_monthly_performances_dataframe(self):
        return self.get_performances_dataframe("ME")

    @property
    def df(self):
        return self._sdf
